# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""A module to acquire tokens from AAD.
"""

import os
import sys
import time
from datetime import timedelta, datetime
from urllib.parse import urlparse
# import webbrowser
import json
from base64 import urlsafe_b64decode


import dateutil.parser
from adal import AuthenticationContext
from adal.constants import TokenResponseFields, OAuth2DeviceCodeResponseParameters


from .dependencies import Dependencies
from .my_utils import single_quote
from .constants import Cloud
from .log import logger
from .display import Display
from .constants import ConnStrKeys
from .adal_token_cache import AdalTokenCache
from .exceptions import KqlEngineError
from .parser import Parser
from .email_notification import EmailNotification
from .aad_helper import AadHelper
from .os_dependent_api import OsDependentAPI


class AuthenticationError(Exception):
    """Raised when authentication fails."""

    def __init__(self, exception, **kwargs):
        super(AuthenticationError, self).__init__()
        exception = exception.exception if isinstance(exception, AuthenticationError) else exception
        self.authentication_method = kwargs.get("authentication_method")
        self.authority = kwargs.get("authority")
        self.resource = kwargs.get("resource")
        self.exception = exception
        self.kwargs = kwargs

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"AuthenticationError('{self.authentication_method}', '{repr(self.exception)}', '{self.kwargs}')"


class ConnKeysKCSB(object):
    """
    Object like dict, every dict[key] can be visited by dict.key
    """

    def __init__(self, conn_kv, data_source):
        self.conn_kv = conn_kv
        self.data_source = data_source
        self.translate_map = {
            "authority_id":                       ConnStrKeys.TENANT,
            "aad_url":                            ConnStrKeys.AAD_URL,
            "aad_user_id":                        ConnStrKeys.USERNAME,
            "password":                           ConnStrKeys.PASSWORD,
            "application_client_id":              ConnStrKeys.CLIENTID,
            "application_key":                    ConnStrKeys.CLIENTSECRET,
            "application_certificate":            ConnStrKeys.CERTIFICATE,
            "application_certificate_thumbprint": ConnStrKeys.CERTIFICATE_THUMBPRINT,
        }


    def __getattr__(self, kcsb_attr_name):
        if kcsb_attr_name == "data_source":
            return self.data_source
        key = self.translate_map.get(kcsb_attr_name)
        return self.conn_kv.get(key)


class AuthenticationMethod(object):
    """Represnting all authentication methods available in Azure Monitor with Python."""

    aad_username_password = "aad_username_password"
    aad_application_key = "aad_application_key"
    aad_application_certificate = "aad_application_certificate"
    aad_code_login = "aad_code_login"

    # external tokens
    azcli_login = "azcli_login"
    azcli_login_by_profile = "azcli_login_by_profile"
    azcli_login_subscription = "azcli_login_subscription"
    managed_service_identity = "managed_service_identity"
    aux_token = "token"


_CLOUD_AAD_URLS = {
    Cloud.PUBLIC:      "https://login.microsoftonline.com",
    Cloud.MOONCAKE:    "https://login.partner.microsoftonline.cn",  # === 'login.chinacloudapi.cn'?
    Cloud.FAIRFAX:     "https://login.microsoftonline.us",
    Cloud.BLACKFOREST: "https://login.microsoftonline.de",
}
_CLOUD_AAD_URLS[Cloud.CHINA] = _CLOUD_AAD_URLS[Cloud.MOONCAKE]
_CLOUD_AAD_URLS[Cloud.GOVERNMENT] = _CLOUD_AAD_URLS[Cloud.FAIRFAX]
_CLOUD_AAD_URLS[Cloud.GERMANY] = _CLOUD_AAD_URLS[Cloud.BLACKFOREST]


_CLOUD_DSTS_AAD_DOMAINS = {
    # Define dSTS domains whitelist based on its Supported Environments & National Clouds list here
    # https://microsoft.sharepoint.com/teams/AzureSecurityCompliance/Security/SitePages/dSTS%20Fundamentals.aspx
    Cloud.PUBLIC:       'dsts.core.windows.net',
    Cloud.MOONCAKE:     'dsts.core.chinacloudapi.cn',  
    Cloud.BLACKFOREST:  'dsts.core.cloudapi.de', 
    Cloud.FAIRFAX:      'dsts.core.usgovcloudapi.net'
}
_CLOUD_DSTS_AAD_DOMAINS[Cloud.CHINA] = _CLOUD_DSTS_AAD_DOMAINS[Cloud.MOONCAKE]
_CLOUD_DSTS_AAD_DOMAINS[Cloud.GOVERNMENT] = _CLOUD_DSTS_AAD_DOMAINS[Cloud.FAIRFAX]
_CLOUD_DSTS_AAD_DOMAINS[Cloud.GERMANY] = _CLOUD_DSTS_AAD_DOMAINS[Cloud.BLACKFOREST]


# shared not cached context per authority
global_adal_context = {}

# shared cached context per authority
global_adal_context_sso = {}


class OAuth2TokenFields(object):
    # taken from here: https://docs.microsoft.com/en-us/azure/app-service/overview-managed-identity?tabs=dotnet

    # The requested access token. The calling web service can use this token to authenticate to the receiving web service.
    ACCESS_TOKEN = 'access_token'

    # The client ID of the identity that was used.
    CLIENT_ID = 'client_id'

    # The timespan when the access token expires. The date is represented as the number of seconds from "1970-01-01T0:0:0Z UTC" (corresponds to the token's exp claim).
    EXPIRES_ON = 'expires_on'

    # The timespan when the access token takes effect, and can be accepted.
    # The date is represented as the number of seconds from "1970-01-01T0:0:0Z UTC" (corresponds to the token's nbf claim).
    NOT_BEFORE = 'not_before'

    # The resource the access token was requested for, which matches the resource query string parameter of the request.
    RESOURCE = 'resource'

    # Indicates the token type value. The only type that Azure AD supports is FBearer. For more information about bearer tokens, 
    # see The OAuth 2.0 Authorization Framework: Bearer Token Usage (RFC 6750).
    TOKEN_TYPE = 'token_type'

    # optional
    ID_TOKEN = 'id_token'
    REFRESH_TOKEN = 'refresh_token'


class _MyAadHelper(AadHelper):

    def __init__(self, kcsb, default_clientid, adal_context=None, adal_context_sso=None, http_client=None, **options):
        global global_adal_context
        global global_adal_context_sso

        super(_MyAadHelper, self).__init__(kcsb, default_clientid, adal_context, adal_context_sso, **options)
        self._http_client = http_client
        # to provide stickiness, to avoid switching tokens when not required
        self._current_token = None
        self._current_adal_context = None
        self._current_authentication_method = None
        self._token_claims_cache = (None, None)

        # options are freezed for authentication when object is created, 
        # to eliminate the need to specify auth option on each query, and to modify behavior on exah query
        self._options = {**options}

        # track warning to avoid repeating
        self._displayed_warnings = []

        url = urlparse(kcsb.data_source)
        self._resource = f"{url.scheme}://{url.hostname}"

        self._authority = kcsb.authority_id or "common"

        self._aad_login_url = self._get_aad_login_url(kcsb.conn_kv.get(ConnStrKeys.AAD_URL))

        self._set_adal_context(adal_context=adal_context, adal_context_sso=adal_context_sso)


        self._client_id = kcsb.application_client_id or default_clientid

        self._username = None
        if all([kcsb.aad_user_id, kcsb.password]):
            self._authentication_method = AuthenticationMethod.aad_username_password
            self._username = kcsb.aad_user_id
            self._password = kcsb.password
        elif all([kcsb.application_client_id, kcsb.application_key]):
            self._authentication_method = AuthenticationMethod.aad_application_key
            self._client_secret = kcsb.application_key
        elif all([kcsb.application_client_id, kcsb.application_certificate, kcsb.application_certificate_thumbprint]):
            self._authentication_method = AuthenticationMethod.aad_application_certificate
            self._certificate = kcsb.application_certificate
            self._thumbprint = kcsb.application_certificate_thumbprint
        else:
            self._authentication_method = AuthenticationMethod.aad_code_login
            self._username = kcsb.aad_user_id  # optional


    # collect this information, in case bug report will be generated
    def get_details(self):
        details = {
            "resource": self._resource,
            "authentication_method": self._current_authentication_method,
            "authority_uri": self._authority_uri,
            "client_id": self._client_id,
            "aad_login_url": self._aad_login_url}
        if (self._current_authentication_method == AuthenticationMethod.aad_username_password
                or self._current_authentication_method == AuthenticationMethod.aad_code_login):
            details["username"] = self._username

        if self._current_token:
            details["resource"] = self._get_token_resource(self._current_token) or self._get_resources_from_token(self._current_token) or details["resource"]
            details["authority_uri"] = self._get_token_authority(self._current_token) or self._get_authority_from_token(self._current_token) or details["authority_uri"]
            details["client_id"] = self._get_token_client_id(self._current_token) or self._get_client_id_from_token(self._current_token) or details["client_id"]
            details["username"] = self._get_token_user_id(self._current_token) or self._get_username_from_token(self._current_token) or details["username"]

        if self._current_authentication_method == AuthenticationMethod.azcli_login_subscription:
            details["subscription"] = self._options.get("try_azcli_login_subscription")
        elif self._current_authentication_method == AuthenticationMethod.azcli_login:
            details["tenant"] = self._authority
        elif self._current_authentication_method == AuthenticationMethod.azcli_login_by_profile:
            details["tenant"] = self._authority
        elif self._current_authentication_method == AuthenticationMethod.managed_service_identity:
            details["msi_params"] = self._options.get("try_msi")
        elif self._current_authentication_method == AuthenticationMethod.aad_application_certificate:
            details["thumbprint"] = self._thumbprint
        elif self._current_authentication_method == AuthenticationMethod.aad_application_key:
            details["client_secret"] = '*****' if self._client_secret else 'NOT-SET'
        elif self._current_authentication_method == AuthenticationMethod.aad_username_password:
            details["password"] = '*****' if self._password else 'NOT-SET'

        return details


    def acquire_token(self):
        """Acquire tokens from AAD."""
        previous_token = self._current_token
        try:
            if self._current_token is not None:
                self._current_token = self._validate_and_refresh_token(self._current_token)

            if self._current_token is None:
                self._current_authentication_method = None
                self._current_adal_context = None

            if self._current_token is None:
                if self._options.get("try_token") is not None:
                    token = self._get_aux_token(token=self._options.get("try_token"))
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._options.get("try_msi") is not None:
                    token = self._get_msi_token(msi_params=self._options.get("try_msi"))
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._options.get("try_azcli_login_subscription") is not None:
                    token = self._get_azcli_token(subscription=self._options.get("try_azcli_login_subscription"))
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._options.get("try_azcli_login"):
                    token = self._get_azcli_token()
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._options.get("try_azcli_login_by_profile"):
                    token = self._get_azcli_token()
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._adal_context_sso is not None:
                    token = self._get_adal_sso_token()
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                if self._adal_context is not None:
                    token = self._get_adal_token()
                    self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                token = None
                self._current_authentication_method = self._authentication_method
                self._current_adal_context = self._adal_context_sso or self._adal_context

                if self._authentication_method is AuthenticationMethod.aad_username_password:
                    logger().debug(f"_MyAadHelper::acquire_token - aad/user-password - resource: '{self._resource}', username: '{self._username}', password: '...', client: '{self._client_id}'")
                    token = self._current_adal_context.acquire_token_with_username_password(self._resource, self._username, self._password, self._client_id)

                elif self._authentication_method is AuthenticationMethod.aad_application_key:
                    logger().debug(f"_MyAadHelper::acquire_token - aad/client-secret - resource: '{self._resource}', client: '{self._client_id}', secret: '...'")
                    token = self._current_adal_context.acquire_token_with_client_credentials(self._resource, self._client_id, self._client_secret)

                elif self._authentication_method is AuthenticationMethod.aad_application_certificate:
                    logger().debug(f"_MyAadHelper::acquire_token - aad/client-certificate - resource: '{self._resource}', client: '{self._client_id}', _certificate: '...', thumbprint: '{self._thumbprint}'")
                    token = self._current_adal_context.acquire_token_with_client_certificate(self._resource, self._client_id, self._certificate, self._thumbprint)                

                elif self._authentication_method is AuthenticationMethod.aad_code_login:
                    logger().debug(f"_MyAadHelper::acquire_token - aad/code - resource: '{self._resource}', client: '{self._client_id}'")
                    code: dict = self._current_adal_context.acquire_user_code(self._resource, self._client_id)
                    url = code[OAuth2DeviceCodeResponseParameters.VERIFICATION_URL]
                    device_code = code[OAuth2DeviceCodeResponseParameters.USER_CODE].strip()

                    device_code_login_notification = self._options.get("device_code_login_notification")
                    if device_code_login_notification == "auto":
                        if self._options.get("notebook_app") in ["azuredatastudiosaw"]:
                            device_code_login_notification = "terminal"
                        elif self._options.get("notebook_app") in ["ipython"]:
                            device_code_login_notification = "popup_interaction"
                        elif self._options.get("notebook_app") in ["visualstudiocode", "azuredatastudio"]:
                            device_code_login_notification = "popup_interaction"
                        elif self._options.get("notebook_app") in ["nteract"]:

                            if self._options.get("kernel_location") == "local":
                                # ntreact cannot execute authentication script, workaround using temp_file_server webbrowser
                                if self._options.get("temp_files_server_address") is not None:
                                    import urllib.parse
                                    indirect_url = f'{self._options.get("temp_files_server_address")}/webbrowser?url={urllib.parse.quote(url)}&kernelid={self._options.get("kernel_id")}'
                                    url = indirect_url
                                    device_code_login_notification = "popup_interaction"
                                else:
                                    device_code_login_notification = "browser"
                            else:
                                device_code_login_notification = "terminal"
                        else:
                            device_code_login_notification = "button"

                    if (self._options.get("kernel_location") == "local"
                            or device_code_login_notification in ["browser"]
                            or (device_code_login_notification == "popup_interaction" and self._options.get("popup_interaction") == "webbrowser_open_at_kernel")):
                        # copy code to local clipboard
                        try:
                            pyperclip = Dependencies.get_module("pyperclip", dont_throw=True)
                            if pyperclip is not None:
                                pyperclip.copy(device_code)
                        except:
                            pass

                    # if  self._options.get("notebook_app")=="papermill" and self._options.get("login_code_destination") =="browser":
                    #     raise Exception("error: using papermill without an email specified is not supported")
                    if device_code_login_notification == "email":
                        params = Parser.parse_and_get_kv_string(self._options.get('device_code_notification_email'), {})
                        email_notification = EmailNotification(**params)
                        subject = f"Kqlmagic device_code {device_code} authentication (context: {email_notification.context})"
                        resource = self._resource.replace("://", ":// ")  # just to make sure it won't be replace in email by safelinks
                        email_message = f"Device_code: {device_code}\n\nYou are asked to authorize access to resource: " \
                                        f"{resource}\n\nOpen the page {url} and enter the code {device_code} to authenticate\n\nKqlmagic"
                        email_notification.send_email(subject, email_message)
                        info_message =f"An email was sent to {email_notification.send_to} with device_code {device_code} to authenticate"
                        Display.showInfoMessage(info_message, display_handler_name='acquire_token', **self._options)

                    elif device_code_login_notification == "browser":
                        # this print is not for debug
                        print(code[OAuth2DeviceCodeResponseParameters.MESSAGE])
                        OsDependentAPI.webbrowser_open(code[OAuth2DeviceCodeResponseParameters.VERIFICATION_URL])

                    elif device_code_login_notification == "terminal":
                        # this print is not for debug
                        print(code[OAuth2DeviceCodeResponseParameters.MESSAGE])
                        sys.stdout.flush()  # Some terminal needs this to ensure the message is shown

                        # Ideally you should wait here, in order to save some unnecessary polling
                        # TODO: add flag to prompt
                        # input("Press Enter after signing in from another device to proceed, CTRL+C to abort.")

                    elif device_code_login_notification == "popup_interaction" and self._options.get("popup_interaction") != "memory_button":
                        before_text = f"<b>{device_code}</b>"
                        button_text = "Copy code to clipboard and authenticate"
                        # before_text = f"Copy code: {device_code} to verification url: {url} and "
                        # button_text='authenticate'
                        # Display.showInfoMessage(f"Copy code: {device_code} to verification url: {url} and authenticate", display_handler_name='acquire_token', **options)
                        Display.show_window(
                            'verification_url',
                            url,
                            button_text=button_text,
                            # palette=Display.info_style,
                            before_text=before_text,
                            display_handler_name='acquire_token',
                            **self._options
                        )

                    else:  # device_code_login_notification == "button":
                        html_str = (
                            f"""<!DOCTYPE html>
                            <html><body>

                            <!-- h1 id="user_code_p"><b>{device_code}</b><br></h1-->

                            <input  id="kql_MagicCodeAuthInput" type="text" readonly style="font-weight: bold; border: none;" size={single_quote(len(device_code))} value={single_quote(device_code)}>

                            <button id='kql_MagicCodeAuth_button', onclick="this.style.visibility='hidden';kql_MagicCodeAuthFunction()">Copy code to clipboard and authenticate</button>

                            <script>
                            var kql_MagicUserCodeAuthWindow = null;
                            function kql_MagicCodeAuthFunction() {{
                                /* Get the text field */
                                var copyText = document.getElementById("kql_MagicCodeAuthInput");

                                /* Select the text field */
                                copyText.select();

                                /* Copy the text inside the text field */
                                document.execCommand("copy");

                                /* Alert the copied text */
                                // alert("Copied the text: " + copyText.value);

                                var w = screen.width / 2;
                                var h = screen.height / 2;
                                params = 'width='+w+',height='+h
                                kql_MagicUserCodeAuthWindow = window.open('{url}', 'kql_MagicUserCodeAuthWindow', params);

                                // TODO: save selected cell index, so that the clear will be done on the lince cell
                            }}
                            </script>

                            </body></html>"""
                        )
                        Display.show_html(html_str, display_handler_name='acquire_token', **self._options)

                    try:
                        token = self._current_adal_context.acquire_token_with_device_code(self._resource, code, self._client_id)
                        logger().debug(f"_MyAadHelper::acquire_token - got token - resource: '{self._resource}', client: '{self._client_id}', token type: '{type(token)}'")
                        self._username = self._username or token.get(TokenResponseFields.USER_ID)

                    finally:
                        html_str = """<!DOCTYPE html>
                            <html><body><script>

                                // close authentication window
                                if (kql_MagicUserCodeAuthWindow && kql_MagicUserCodeAuthWindow.opener != null && !kql_MagicUserCodeAuthWindow.closed) {
                                    kql_MagicUserCodeAuthWindow.close()
                                }
                                // TODO: make sure, you clear the right cell. BTW, not sure it is a must to do any clearing

                                // clear output cell
                                Jupyter.notebook.clear_output(Jupyter.notebook.get_selected_index())

                                // TODO: if in run all mode, move to last cell, otherwise move to next cell
                                // move to next cell

                            </script></body></html>"""

                        Display.show_html(html_str, display_handler_name='acquire_token', **self._options)

                self._current_token = self._validate_and_refresh_token(token)

            if self._current_token is None:
                raise AuthenticationError("No valid token.")

            if self._current_token != previous_token:
                self._warn_token_diff_from_conn_str()
            else:
                logger().debug(f"_MyAadHelper::acquire_token - valid token exist - resource: '{self._resource}', username: '{self._username}', client: '{self._client_id}'")

            return self._create_authorization_header()
        except Exception as e:
            kwargs = self._get_authentication_error_kwargs()
            raise AuthenticationError(e, **kwargs)


    #
    # Assume OAuth2 format (e.g. MSI Token) too
    #
    def _get_token_access_token(self, token:dict, default_access_token:str=None)->str:
        return token.get(TokenResponseFields.ACCESS_TOKEN) or token.get(OAuth2TokenFields.ACCESS_TOKEN) or default_access_token


    def _get_token_client_id(self, token:dict, default_client_id:str=None)->str:
        return token.get(TokenResponseFields._CLIENT_ID) or token.get(OAuth2TokenFields.CLIENT_ID) or default_client_id


    def _get_token_expires_on(self, token:dict, default_expires_on:str=None)->str:
        expires_on = default_expires_on
        if token.get(TokenResponseFields.EXPIRES_ON) is not None:
            expires_on = token.get(TokenResponseFields.EXPIRES_ON)
        elif token.get(OAuth2TokenFields.EXPIRES_ON) is not None:
            # The date is represented as the number of seconds from "1970-01-01T0:0:0Z UTC" (corresponds to the token's exp claim).
            expires_on = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(token.get(OAuth2TokenFields.EXPIRES_ON)))
        return expires_on


    def _get_token_not_before(self, token:dict, default_not_before:str=None)->str:
        not_before = default_not_before
        if token.get(OAuth2TokenFields.NOT_BEFORE) is not None:
            # The date is represented as the number of seconds from "1970-01-01T0:0:0Z UTC" (corresponds to the token's nbf claim).
            not_before = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(token.get(OAuth2TokenFields.NOT_BEFORE)))
        return not_before


    def _get_token_token_type(self, token:dict, default_token_type:str=None)->str:
        return token.get(TokenResponseFields.TOKEN_TYPE) or token.get(OAuth2TokenFields.TOKEN_TYPE) or default_token_type


    def _get_token_resource(self, token:dict, default_resource:str=None)->str:
        return token.get(TokenResponseFields.RESOURCE) or token.get(OAuth2TokenFields.RESOURCE) or default_resource


    def _get_token_user_id(self, token:dict, default_user_id:str=None)->str:
        return token.get(TokenResponseFields.USER_ID) or default_user_id


    def _get_token_refresh_token(self, token:dict, default_refresh_token:str=None)->str:
        return token.get(TokenResponseFields.REFRESH_TOKEN) or token.get(OAuth2TokenFields.REFRESH_TOKEN) or default_refresh_token


    def _get_token_id_token(self, token:dict, default_id_token:str=None)->str:
        return token.get(OAuth2TokenFields.ID_TOKEN) or default_id_token


    def _get_token_authority(self, token:dict, default_authority:str=None)->str:
        return token.get(TokenResponseFields._AUTHORITY) or default_authority


    def _create_authorization_header(self)->str:
        "create content for http authorization header"
        access_token = self._get_token_access_token(self._current_token)
        if access_token is None:
            raise AuthenticationError("Not a valid token, property 'access_token' is not present.")

        token_type = self._get_token_token_type(self._current_token)
        if token_type is None:
            raise AuthenticationError("Unable to determine the token type. Neither 'tokenType' nor 'token_type' property is present.")

        return f"{token_type} {access_token}"


    def _get_username_from_token(self, token:dict)->str:
        "retrieves username from in id token or access token claims"

        id_token_claims = self._get_token_claims(self._get_token_id_token(token))
        access_token_claims = self._get_token_claims(self._get_token_access_token(token))
        claims = {**access_token_claims, **id_token_claims}
        username = claims.get("preferred_username") or claims.get("unique_name") or claims.get("upn") or claims.get("email") or claims.get("sub")
        return username


    def _get_expires_on_from_token(self, token:dict)->str:
        "retrieve expires_on from access token claims"
        expires_on = None
        claims = self._get_token_claims(self._get_token_access_token(token))
        exp = claims.get("exp")
        if exp is not None:
            expires_on = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(exp))
        return expires_on


    def _get_not_before_from_token(self, token:dict)->str:
        "retrieve not_before from access token claims"
        not_before = None
        claims = self._get_token_claims(self._get_token_access_token(token))
        nbf = claims.get("nbf")
        if nbf is not None:
            not_before = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(nbf))
        return not_before


    def _get_client_id_from_token(self, token:dict)->str:
        "retrieve client_id from access token claims"
        claims = self._get_token_claims(self._get_token_access_token(token))
        client_id = claims.get("client_id") or claims.get("appid") or claims.get("azp")
        return client_id

    
    def _get_resources_from_token(self, token:dict)->list:
        "retrieve resource list from access token claims"
        resources = None
        claims = self._get_token_claims(self._get_token_access_token(token))
        resources = claims.get("aud")
        if type(resources) == str:
            resources = [resources]
        return resources


    def _get_authority_from_token(self, token:dict)->str:
        "retrieve authority_uri from access token claims"
        authority_uri = None
        try:
            claims = self._get_token_claims(self._get_token_access_token(token))
            tenant_id = claims.get("tid")
            issuer = claims.get("iss")

            if tenant_id is None and issuer is not None and issuer.startswith("http"):
                from urllib.parse import urlparse
                url_obj = urlparse(issuer)
                tenant_id = url_obj.path

            if tenant_id is not None:
                if tenant_id.startswith("http"):
                    authority_uri = tenant_id
                else:
                    if tenant_id.startswith("/"):
                        tenant_id = tenant_id[1:]
                    if tenant_id.endswith("/"):
                        tenant_id = tenant_id[:-1]
                    authority_uri = f"{self._aad_login_url}/{tenant_id}"
        except:
            pass

        return authority_uri


    def _get_adal_token(self)->str:
        "retrieve token from adal cache"
        token = None
        self._current_authentication_method = self._authentication_method
        try:
            self._current_adal_context = self._adal_context
            token = self._current_adal_context.acquire_token(self._resource, self._username, self._client_id)
        except:
            pass
        logger().debug(f"_MyAadHelper::_get_adal_token {'failed' if token is None else 'succeeded'} to get token")
        return token

                
    def _get_adal_sso_token(self)->str:
        "retrieve token from adal sso cache"
        token = None
        self._current_authentication_method = self._authentication_method
        try:
            self._current_adal_context = self._adal_context_sso
            token = self._current_adal_context.acquire_token(self._resource, self._username, self._client_id)
        except:
            pass
        logger().debug(f"_MyAadHelper::_get_adal_sso_token {'failed' if token is None else 'succeeded'} to get token")
        return token


    def _get_aux_token(self, token:dict)->str:
        "retrieve token from aux token"
        self._current_authentication_method = AuthenticationMethod.aux_token
        try:
            token = token
        except:
            pass
        logger().debug(f"_MyAadHelper::_get_aux_token {'failed' if token is None else 'succeeded'} to get token")
        return token


    def _get_azcli_token(self, subscription:str=None)->str:
        "retrieve token from azcli login"
        token = None
        tenant = self._authority if subscription is None else None
        self._current_authentication_method = self._current_authentication_method = AuthenticationMethod.azcli_login_subscription if subscription is not None else AuthenticationMethod.azcli_login
        try:
            # requires azure-cli-core to be installed
            # from azure.cli.core._profile import _CLIENT_ID as AZCLI_CLIENT_ID
            from azure.common.credentials import get_cli_profile  # pylint: disable=no-name-in-module, import-error
            profile = get_cli_profile()
            credential, _subscription, _tenant = profile.get_raw_token(resource=self._resource, subscription=subscription, tenant=tenant)
            token_type, access_token, token = credential  # pylint: disable=unused-variable
        except:
            pass
        logger().debug(f"_MyAadHelper::_get_azcli_token {'failed' if token is None else 'succeeded'} to get token - subscription: '{subscription}', tenant: '{tenant}'")
        return token


    def _get_msi_token(self, msi_params=None)->str:
        "retrieve token from managed service identity"
        msi_params = msi_params or {}
        token = None
        self._current_authentication_method = AuthenticationMethod.managed_service_identity
        try:
            from msrestazure.azure_active_directory import MSIAuthentication
            # allow msi_params to overrite the connection string resource
            credentials = MSIAuthentication(**{"resource":self._resource, **msi_params})
            token = credentials.token
        except:
            pass

        logger().debug(f"_MyAadHelper::_get_msi_token {'failed' if token is None else 'succeeded'} to get token - msi_params: '{msi_params}'")
        return token


    def _validate_and_refresh_token(self, token:dict)->str:
        "validate token is valid to use now. Now is between not_before and expires_on. If exipred try to refresh"
        valid_token = None
        if token is not None:
            resource = self._get_token_resource(token) or self._resource
            not_before = self._get_token_not_before(token) or self._get_not_before_from_token(token)
            if not_before is not None:
                not_before_datetime = dateutil.parser.parse(not_before)
                current_datetime = datetime.now() - timedelta(minutes=1)
                if not_before_datetime > current_datetime:
                    logger().debug(f"_MyAadHelper::_validate_and_refresh_token - failed - token can be used not before {not_before} - resource: '{resource}'")
                    self._warn_on_token_validation_failure(f"access token cannot be used before {not_before}")
                    return None

            expires_on = self._get_token_expires_on(token) or self._get_expires_on_from_token(token)
            if expires_on is not None:
                expiration_datetime = dateutil.parser.parse(expires_on)
            else:
                expiration_datetime = datetime.now() + timedelta(minutes=30)

            current_datetime = datetime.now() + timedelta(minutes=1)
            if expiration_datetime > current_datetime:
                valid_token = token
                logger().debug(f"_MyAadHelper::_validate_and_refresh_token - succeeded, no need to refresh yet, expires on {expires_on} - resource: '{resource}'")
            else:
                logger().debug(f"_MyAadHelper::_validate_and_refresh_token - token expires on {expires_on} need to refresh - resource: '{resource}'")
                refresh_token = self._get_token_refresh_token(token)
                if refresh_token is not None:
                    try:
                        if self._current_adal_context is None:
                            authority_uri = self._get_token_authority(token) or self._get_authority_from_token(token) or self._authority_uri
                            self._current_adal_context = AuthenticationContext(authority_uri, cache=None)
                        client_id = self._get_token_client_id(token) or self._get_client_id_from_token(token) or self._client_id
                        valid_token = self._current_adal_context.acquire_token_with_refresh_token(refresh_token, client_id, resource)
                    except Exception as e:
                        self._warn_on_token_validation_failure(f"access token expired on {expires_on}, failed to refresh access token, Exception: {e}")

                    logger().debug(f"_MyAadHelper::_validate_and_refresh_token - {'failed' if token is None else 'succeeded'} to refresh token - resource: '{resource}'")
                else:
                    logger().debug(f"_MyAadHelper::_validate_and_refresh_token - failed to refresh expired token, token doesn't contain refresh token - resource: '{resource}'")
                    self._warn_on_token_validation_failure(f"access token expired on {expires_on}, and token entry has no refresh_token")

        return valid_token


    def _set_adal_context(self, adal_context=None, adal_context_sso=None):
        "set the adal context"
        self._authority_uri = f"{self._aad_login_url}/{self._authority}"

        self._adal_context = adal_context
        if self._adal_context is None:
            if global_adal_context.get(self._authority_uri) is None:
                global_adal_context[self._authority_uri] = AuthenticationContext(self._authority_uri, cache=None)
            self._adal_context = global_adal_context.get(self._authority_uri)

        self._adal_context_sso = None
        if self._options.get("enable_sso"):
            self._adal_context_sso = adal_context_sso
            if self._adal_context_sso is None:
                if global_adal_context_sso.get(self._authority_uri) is None:
                    cache = AdalTokenCache.get_cache(self._authority_uri, **self._options)
                    if cache is not None:
                        global_adal_context_sso[self._authority_uri] = AuthenticationContext(self._authority_uri, cache=cache)
                self._adal_context_sso = global_adal_context_sso.get(self._authority_uri)


    def _get_aad_login_url(self, aad_login_url=None):
        if aad_login_url is None:
            cloud = self._options.get("cloud")
            aad_login_url = _CLOUD_AAD_URLS.get(cloud)
            if aad_login_url is None:
                raise KqlEngineError(f"AAD is not known for this cloud '{cloud}', please use aadurl property in connection string.")
        return aad_login_url


    def _warn_on_token_validation_failure(self, message)->None:
        if self._options.get("auth_token_warnings"):
            if self._current_authentication_method is not None and message is not None:
                warn_message =f"Can't use '{self._current_authentication_method}' token entry, {message}'"
                Display.showWarningMessage(warn_message, display_handler_name='acquire_token', **self._options)


    def _warn_token_diff_from_conn_str(self)->None:
        if self._options.get("auth_token_warnings"):
            token = self._current_token
            if token is not None:
                # to avoid more than one warning per connection, keep track of already displayed warnings
                access_token = self._get_token_access_token(token)
                key = hash((access_token))
                if key in self._displayed_warnings:
                    return
                else:
                    self._displayed_warnings.append(key)

                token_username = self._get_token_user_id(token) or self._get_username_from_token(token)
                if token_username is not None and self._username is not None and token_username != self._username:
                    warn_message =f"authenticated username '{token_username}' is different from connectiion string username '{self._username}'"
                    Display.showWarningMessage(warn_message, display_handler_name='acquire_token', **self._options)

                token_authority_uri = self._get_token_authority(token) or self._get_authority_from_token(token)
                if token_authority_uri != self._authority_uri and not self._authority_uri.endswith("/common") and not token_authority_uri.endswith("/common"):
                    warn_message =f"authenticated authority '{token_authority_uri}' is different from connectiion string authority '{self._authority_uri}'"
                    Display.showWarningMessage(warn_message, display_handler_name='acquire_token', **self._options)

                token_client_id = self._get_token_client_id(token) or self._get_client_id_from_token(token)
                if token_client_id is not None and self._client_id is not None and token_client_id != self._client_id:
                    warn_message =f"authenticated client_id '{token_client_id}' is different from connectiion string client_id '{self._client_id}'"
                    Display.showWarningMessage(warn_message, display_handler_name='acquire_token', **self._options)

                token_resources = self._get_token_resource(token) or self._get_resources_from_token(token)
                if type(token_resources) == str:
                    token_resources = [token_resources]
                if token_resources is not None and self._resource is not None and self._resource not in token_resources:
                    warn_message =f"authenticated resources '{token_resources}' does not include connectiion string resource '{self._resource}'"
                    Display.showWarningMessage(warn_message, display_handler_name='acquire_token', **self._options)


    def _get_authentication_error_kwargs(self):
        " collect info for AuthenticationError exception and raise it"
        kwargs = {}
        if self._current_authentication_method is AuthenticationMethod.aad_username_password:
            kwargs = {"username": self._username, "client_id": self._client_id}
        elif self._current_authentication_method is AuthenticationMethod.aad_application_key:
            kwargs = {"client_id": self._client_id}
        elif self._current_authentication_method is AuthenticationMethod.aad_code_login:
            kwargs = {"client_id": self._client_id}
        elif self._current_authentication_method is AuthenticationMethod.aad_application_certificate:
            kwargs = {"client_id": self._client_id, "thumbprint": self._thumbprint}
        elif self._current_authentication_method is AuthenticationMethod.managed_service_identity:
            kwargs = self._options.get("try_msi")
        elif self._current_authentication_method is AuthenticationMethod.azcli_login:
            pass
        elif self._current_authentication_method is AuthenticationMethod.azcli_login_by_profile:
            pass
        elif self._current_authentication_method is AuthenticationMethod.azcli_login_subscription:
            kwargs = {"subscription": self._options.get("try_azcli_login_subscription")}
        elif self._current_authentication_method is AuthenticationMethod.aux_token:
            token_dict = {}
            for key in self._options.get("try_token"):
                if key in [TokenResponseFields.ACCESS_TOKEN, OAuth2TokenFields.ACCESS_TOKEN, TokenResponseFields.REFRESH_TOKEN, OAuth2TokenFields.REFRESH_TOKEN, OAuth2TokenFields.ID_TOKEN]:
                    token_dict[key] = f"..."
                else:
                    token_dict[key] = self._options.get("try_token")[key]
            kwargs = token_dict
        else:
            pass

        authority = None
        if self._current_adal_context is not None:
            authority = self._current_adal_context.authority.url
        elif self._current_authentication_method == self._authentication_method:
            authority =  self._authority_uri
        elif self._current_token is not None:
            authority = self._get_authority_from_token(self._current_token)
        else:
            authority = authority or self._current_authentication_method

        if authority is None:
            if self._current_authentication_method in [AuthenticationMethod.managed_service_identity, AuthenticationMethod.azcli_login_subscription, AuthenticationMethod.aux_token]:
                authority = self._current_authentication_method
            else:
                authority = self._authority_uri

        kwargs["authority"] = authority
        kwargs["authentication_method"] = self._current_authentication_method
        kwargs["resource"] = self._resource

        return kwargs


    def _get_token_claims(self, jwt_token:str)->dict:
        "get the claims from the token. To optimize it caches the last token/claims"
        
        claims_token, claims = self._token_claims_cache
        if jwt_token == claims_token:
            return claims
        claims = {}
        if jwt_token:
            try:
                base64_header, base64_claims, _ = jwt_token.split('.')  # pylint: disable=unused-variable
                json_claims = self.base64url_decode(base64_claims)
                claims = json.loads(json_claims)
            except Exception as e:
                # this print is not for debug
                print(f"claims error: {e}")
                pass
            self._token_claims_cache = (jwt_token, claims)
        return claims


    def base64url_decode(self, url_str:str)->str:
        size = len(url_str) % 4
        if size != 0:
            padding_size = 4 - size
            if padding_size == 2:
                url_str += '=='
            elif padding_size == 1:
                url_str += '='
            else:
                raise ValueError(f"Invalid base64 url string: {url_str}")
        return urlsafe_b64decode(url_str.encode('utf-8')).decode('utf-8')        


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from datetime import datetime
import json
import string
import random
import threading


from adal.constants import TokenResponseFields


from .my_utils import json_dumps
from .sso_storage import get_sso_store
from .dict_db_storage import DictDbStorage
from .log import logger


def _string_cmp(str1, str2):
    '''Case insensitive comparison. Return true if both are None'''
    str1 = str1 or ''
    str2 = str2 or ''
    return str1.lower() == str2.lower()


# pylint: disable=too-few-public-methods
class AdalTokenCacheKey(object): 

    def __init__(self, authority, resource, client_id, user_id):
        logger().debug(f"AdalTokenCacheKey authority {authority} client_id {client_id}  user_id {user_id} ")
        
        self.authority = authority
        self.resource = resource
        self.client_id = client_id
        self.user_id = user_id


    def __hash__(self):
        return hash((self.authority, self.resource, self.client_id, self.user_id))


    def __eq__(self, other)->bool:
        return (_string_cmp(self.authority, other.authority)
                and _string_cmp(self.resource, other.resource)
                and _string_cmp(self.client_id, other.client_id)
                and _string_cmp(self.user_id, other.user_id))


    def __ne__(self, other):
        return not self == other
# pylint: enable=too-few-public-methods


# pylint: disable=protected-access
def _get_cache_key(entry):
    return AdalTokenCacheKey(
        entry.get(TokenResponseFields._AUTHORITY), 
        entry.get(TokenResponseFields.RESOURCE), 
        entry.get(TokenResponseFields._CLIENT_ID), 
        entry.get(TokenResponseFields.USER_ID))
# pylint: enable=protected-access


class AdalTokenCache(object):

    @classmethod
    def get_cache(cls, cache_selector_key, **options):
        store = get_sso_store(cache_selector_key, **options)
        if store:
            cache = AdalTokenCache(store)
            return cache


    def __init__(self, store:DictDbStorage, state=None):
        self._lock = threading.RLock()

        self._cache = {}
        self._store = store
        if state:
            self.deserialize(state)
        self.has_state_changed = False


    def find(self, query):
        logger().debug(f"AdalTokenCache find(self, query)")

        '''find entries in cache'''
        with self._lock:
            state = self._store.restore()
            self.deserialize(state)
            return self._query_cache(
                query.get(TokenResponseFields.IS_MRRT), 
                query.get(TokenResponseFields.USER_ID), 
                query.get(TokenResponseFields._CLIENT_ID)
            )


    def remove(self, entries):
        logger().debug(f"AdalTokenCache remove(self, query)")

        '''remove entries from cache'''
        with self._lock:
            state = self._store.restore()
            self.deserialize(state)
            removed = None
            for e in entries:
                key = _get_cache_key(e)
                removed = self._cache.pop(key, None) or removed
            if removed:
                self.has_state_changed = True
                state = self.serialize()
                self._store.save(state)


    def add(self, entries):
        logger().debug(f"AdalTokenCache add(self, entries")

        '''add entries to cache'''
        with self._lock:
            state: str = self._store.restore()
            self.deserialize(state)
            added = None
            for e in entries:
                key = _get_cache_key(e)
                self._cache[key] = e
                added = True
            if added:
                self.has_state_changed = True
                state = self.serialize()
                self._store.save(state)


    def _random_string(self) -> str:
        return ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(random.randint(1, 100)))


    def serialize(self) -> str:
        '''serialize cache'''
        with self._lock:
            state_obj = {
                "description": "kqlmagic",
                "version": 1,
                "timestamp": int(datetime.utcnow().timestamp()),
                "random_string": self._random_string(),  # makes length and content different each time
                "cache_values": list(self._cache.values()),
            }
            return json_dumps(state_obj)


    def deserialize(self, state: str):
        '''deserialize cache'''
        with self._lock:
            if state:
                self._cache.clear()
                state_obj = json.loads(state)
                if (state_obj.get("description") == "kqlmagic"
                        and state_obj.get("version") == 1
                        and state_obj.get("timestamp") < int(datetime.utcnow().timestamp())
                        and state_obj.get("random_string") 
                        and len(state_obj.get("random_string")) >= 1 
                        and len(state_obj.get("random_string")) <= 100
                        and state_obj.get("cache_values")):
                    cache_values = state_obj["cache_values"]
                    for val in cache_values:
                        key = _get_cache_key(val)
                        self._cache[key] = val


    def read_items(self):
        '''output list of tuples in (key, authentication-result)'''
        with self._lock:
            state = self._store.restore()
            self.deserialize(state)
            return self._cache.items()


    def _query_cache(self, is_mrrt, user_id, client_id):
        '''query cache for matches'''
        matches = []
        for k in self._cache:
            v = self._cache[k]
            # None value will be taken as wildcard match
            # pylint: disable=too-many-boolean-expressions
            if ((is_mrrt is None or is_mrrt == v.get(TokenResponseFields.IS_MRRT))
                    and (user_id is None or _string_cmp(user_id, v.get(TokenResponseFields.USER_ID))) 
                    and (client_id is None or _string_cmp(client_id, v.get(TokenResponseFields._CLIENT_ID)))):
                matches.append(v)
        return matches



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""A module that manage package version.
"""


from .dependencies import Dependencies
from .constants import Constants, Email, SsoEnvVarParam


_KQL_URL                   = "https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/"
_APPINSIGHTS_URL           = "https://docs.microsoft.com/en-us/azure/application-insights/app-insights-overview?toc=/azure/azure-monitor/toc.json"
_LOGANALYTICS_URL          = "https://docs.microsoft.com/en-us/azure/log-analytics/log-analytics-queries?toc=/azure/azure-monitor/toc.json"
_AZUREMONITOR_URL          = "https://docs.microsoft.com/en-us/azure/azure-monitor/"
_KUSTO_URL                 = "https://docs.microsoft.com/en-us/azure/data-explorer/"
_KQLMAGIC_DOWNLOADS_URL    = "https://pepy.tech/project/Kqlmagic"
_KQLMAGIC_INSTALL_URL      = "https://pypi.org/project/Kqlmagic/"
_KQLMAGIC_README_URL       = "https://github.com/microsoft/jupyter-Kqlmagic/blob/master/README.md"
_KQLMAGIC_GITHUB_URL       = "https://github.com/microsoft/jupyter-Kqlmagic"
_KQLMAGIC_LICENSE_URL      = "https://github.com/microsoft/jupyter-Kqlmagic/blob/master/LICENSE.TXT"
_KQLMAGIC_CONTRIBUTORS_URL = "https://github.com/microsoft/jupyter-Kqlmagic/blob/master/CONTRIBUTORS.md"
_KQLMAGIC_FAQ_URL          = "https://raw.githubusercontent.com/microsoft/jupyter-Kqlmagic/master/FAQ.html"

_NEED_SUPPORT_SECTION = """## Need Support?
- **Have a feature request for Kqlmagic?** Please post it on [User Voice](https://feedback.azure.com/forums/913690-azure-monitor) to help us prioritize
- **Have a technical question?** Ask on [Stack Overflow with tag "Kqlmagic"](https://stackoverflow.com/questions/tagged/Kqlmagic)
- **Need Support?** Every customer with an active Azure subscription has access to [support](https://docs.microsoft.com/en-us/azure/azure-supportability/how-to-create-azure-support-request) with guaranteed response time.  Consider submitting a ticket and get assistance from Microsoft support team
- **Found a bug?** Please help us fix it by thoroughly documenting it and [filing an issue](https://github.com/Microsoft/jupyter-Kqlmagic/issues/new).
"""


_HELP_OPTIONS = ""


_HELP_COMMANDS = """## Overview
Except submitting kql queries, few other commands are included that may help using the Kqlmagic.<br>
- Only one command can be executed per magic transaction.<br>
- A command must start with a double hyphen-minus ```--```<br>
- If command is not specified, the default command ```"submit"``` is assumed, that submits the query.<br>

## Commands
The following commands are supported:<br>
- **submit** - Execute the query and return result. <br>
    - Options can be used to customize the behavior of the transaction.<br>
    - The query can parametrized.<br>
    - This is the default command.<br>
<br>

- **version** - Displays Kqlmagic current version string.<br>
<br>

- **bug_report** - gathers last cell execution state, ready to be copy/paste to Kqlmagic github issues (you must have an account in github)<br>
      - Kqlmagic version<br>
      - platform information<br>
      - python information<br>
      - Kqlmagic dependencies version <br>
      - Kqlmagic default options<br>
      - Kqlmagic connections<br>
      - Kqlmagic environment variables<br>
      - Last kql execution context (options, parameters, tags, request, response), including error stack in case it failed<br>
<br>

- **py** - execute python code in user's namespace, code may change namespace state, and if last statement is an expression it returns value.<br>
<br>

- **pyrw** - execute python code in user's namespace, code may change namespace state, and if last statement is an expression it returns value.<br>
<br>

- **pyro** - execute python code in user's namespace, namespace state does NOT change, and if last statement is an expression it returns value.<br>
<br>

- **line_magic** - executes a python kernel line magic.<br>
    - Only this line will be input to the specified line magic<br>
    - "%" syntax can be used instaed of the command name<br>
<br>

- **cell_magic** - executes a python kernel cell magic.<br>
    - All following lines will be input to the specified cell magic<br>
    - "%%" syntax can be used instead of the command name<br>
<br>

- **activate_kernel** - activates kqlmagic kernel over python kernel.<br>
    - After activation, code cells executes by default as kqlmagic cells<br>
<br>

- **deactivate_kernel** - deactivates kqlmagic kernel over python kernel (switch back to python kernel).<br>
<br>

- **cache** - enables/disables caching query results to the specified folder.<br>
    - OPTIONAL - Stored cached data can be also queried using the cache schema: cache://<cache-folder-name>
<br>

- **use_cache** - enables/disables using query results stored in specified folder as cache.<br>
    - OPTIONAL - Stored cached data can be also queried using the cache schema: cache://<cache-folder-name>
<br>

- **conn** - Displays connections info.<br>
<br>

- **banner** - Displays Kqlmagic init banner.<br>
<br>

- **usage** - Displays usage of Kqlmagic.<br>
<br>

- **config** - get/set Kqlmagic default option<br>
<br>

- **faq** - Displays Frequently Asked Questions on Kqlmagic.<br>
<br>

- **help "topic"** - Displays information about the topic.<br>
    - To get the list of all the topics, execute ```%kql --help "help"```<br>
<br>

- **palette - Display information about the current or other named color palette.<br>
    - The behaviour of this command will change based on the specified option:
    - -palette_name, -palette_colors, palette_reverse, -palette_desaturation, execute ```%kql --palette -palette_name "Reds"```<br>
<br>

- **palettes - Display information about all available palettes.<br>
    - The behaviour of this command will change based on the specified option:
    - -palette_colors, palette_reverse, -palette_desaturation, execute ```%kql --palettes -palette_desaturation 0.75```<br>
<br>

- **schema "database"** - Returns the database schema as a python dict (displayed as a json format). <br>
    - To get Azure Data Explorer database schema: ```%kql --schema "databasename@clustername"```<br>
    - To get application insights app schema: ```%kql --schema "appname@applicationinsights"```<br>
    - To get log analytics workspace schema: ```%kql --schema "workspacename@loganalytics"```<br>
    - To get current connection database schema ```%kql --schema```<br>
    - If -conn option is sepcified it will override the database value.<br>
<br>

- **cache - Enables caching query results to a cache folder, or disbale. <br>
    - To enable caching to folder XXX, execute: ```%kql --cache "XXX"```<br>
    - To disable caching, execute: ```%kql --cache None```<br>
    - Once results are cached, the results can be used by enabling the use of the cache, with the --use_cache command.<br>
<br>

- **use_cache - Enables use of cached results from a cache folder. <br>
    - To enable use of cache from folder XXX, execute: ```%kql --use_cache "XXX"```<br>
    - To disable use of cache, execute: ```%kql --use_cache None```<br>
    - Once enabled, intead of quering the data source, the results are retreived from the cache.<br>
<br>

## Examples:
```%kql --version```<br><br>
```%kql --banner```<br><br>
```%kql --usage```<br><br>
```%kql --config "show_query_time"```<br><br>
```%kql --config "show_query_time = True"```<br><br>
```%kql --config```<br><br>
```%kql --faq```<br><br>
```%kql --help "help"```<br><br>
```%kql --help "options"```<br><br>
```%kql --help "conn"```<br><br>
```%kql --palette -palette_name "Reds"```<br><br>
```%kql --schema 'DEMO_APP@applicationinsights'```<br><br>
```%kql --cache "XXX"```<br><br>
```%kql --use_cache None```<br><br>
```%kql --submit appinsights://appid='DEMO_APP';appkey='DEMO_KEY' pageViews | count```<br><br>
```%kql --palettes -palette_desaturation 0.75```
```%kql pageViews | count```
"""


_USAGE = f"""## Usage:
**line usage:** ```%kql [command] [conn] [result <<] [options] [query]```

**cell usage:** ```%%kql [conn] [result <<] [options] [query] [[EMPTY-LINE [result <<] [options]]*```<br>

- **command** - The command to be executed. 
    - If not specified the query will be submited, if exist.<br>
    - All commands start with a double hyphen-minus ```--```<br>
    - To get more information, execute ```%kql --help "commands"```<br>
<br>

- **conn** - Connection string or reference to a connection to Azure Monitor resource.<br>
    - If not specified the current (last created or used) connection will be used.<br>
    - The conncan be also specified in the options parts, using the option ```-conn```<br>
    - To get more information, execute ```%kql --help "conn"```<br>
<br>

- **result** - Python variable name that will be assigned with the result of the query.<br>
    - Query results are always assigned to ```_``` and to ```_kql_raw_result_``` python variables.<br>
    - If not specified and is last query in the cell, the results will be displayed.<br>
<br>

- **options** - Options that customize the behavior of the Kqlmagic for this transaction only.<br>
    - All options start with a hyphen-minus ```-```<br>
    - To get more information, execute ```%kql --help "options"```<br>
<br>

- **query** - Kusto Query language (kql) query that will be submited to the specified connection or to the current connection.<br>
    - The query can be also sepcified in the options parts, using the option ```-query```<br>
    - To get more information, browse https://docs.microsoft.com/en-us/azure/kusto/query/index 
<br>

## Examples:
```%kql --version```<br><br>
```%kql --usage```<br><br>
```%%kql appinsights://appid='DEMO_APP';appkey='DEMO_KEY' 
    pageViews 
    | where client_City != '' 
    | summarize count() by client_City 
    | sort by count_ 
    | limit 10```<br><br>
```%%kql pageViews | where client_City != '' 
    | summarize count() by client_City | sort by count_ 
    | limit 10```<br><br>
```%kql DEMO_APP@appinsights pageViews | count```

## Get Started Notebooks:

* [Get Started with Kqlmagic for Kusto](https://mybinder.org/v2/gh/Microsoft/jupyter-Kqlmagic/master?filepath=notebooks%2FQuickStart.ipynb)

* [Get Started with Kqlmagic for Application Insights](https://mybinder.org/v2/gh/Microsoft/jupyter-Kqlmagic/master?filepath=notebooks%2FQuickStartAI.ipynb)

* [Get Started with Kqlmagic for Log Analytics](https://mybinder.org/v2/gh/Microsoft/jupyter-Kqlmagic/master?filepath=notebooks%2FQuickStartLA.ipynb)


* [Parametrize your Kqlmagic query with Python](https://mybinder.org/v2/gh/Microsoft/jupyter-Kqlmagic/master?filepath=notebooks%2FParametrizeYourQuery.ipynb)

* [Choose colors palette for your Kqlmagic query chart result](https://mybinder.org/v2/gh/Microsoft/jupyter-Kqlmagic/master?filepath=notebooks%2FColorYourCharts.ipynb)

{_NEED_SUPPORT_SECTION}"""


_HELP_HELP = f"""## Overview
Help command is a tool to get more information on a topics that are relevant to Kqlmagic.<br>help

usage: ```%kql --help "topic"```<br>

## Topics
- **usage** - How to use Kqlmagic.<br>
<br>

- **config** - Lists Kqlmagic default options. The same as --config without parameters.<br>
<br>

- **faq** - Reference to Kqlmagic FAQ<br>
<br>

- **conn** - Lists the available connection string variation, and how their are used to authenticatie to data sources.<br>
<br>

- **query** / **kql** - [Reference to resources Kusto Query language, aka kql, documentation]({_KQL_URL})<br>
<br>

- **options** - Lists the available options, and their behavior impact on the submit query command.<br>
<br>

- **env** - Lists all {Constants.MAGIC_CLASS_NAME_UPPER} environment variables and their purpose.<br>
<br>

- **commands** - Lists the available commands, and what they do.<br>
<br>

- **proxies** - How to use Kqlmagic via proxies.<br>
<br>

- **sql** - Azure Data Explorer SQL support.<br>
<br>

- **aria** - How to query Aria's Azure Data Explorer databases.<br>
<br>

- **client-request-properties** - How to use Client Request properties, and properties list.<br>
<br>

- **request-tags** - How to tag request headers.<br>
<br>

- **AzureMonitor**- [Reference to resources Azure Monitor tools]({_AZUREMONITOR_URL})<br>
Azure Monitor, which now includes Log Analytics and Application Insights, provides sophisticated tools for collecting and analyzing telemetry that allow you to maximize the performance and availability of your cloud and on-premises resources and applications. It helps you understand how your applications are performing and proactively identifies issues affecting them and the resources they depend on.<br>
<br>

- **AzureDataExplorer** / **kusto**- [Reference to resources Azure Data Explorer (kusto) service]({_KUSTO_URL})<br>
Azure Data Explorer is a fast and highly scalable data exploration service for log and telemetry data. It helps you handle the many data streams emitted by modern software, so you can collect, store, and analyze data. Azure Data Explorer is ideal for analyzing large volumes of diverse data from any data source, such as websites, applications, IoT devices, and more.<br>
<br>

- **LogAnalytics**- [Reference to resources Log Analytics service]({_LOGANALYTICS_URL})<br>
Log data collected by Azure Monitor is stored in Log Analytics which collects telemetry and other data from a variety of sources and provides a query language for advanced analytics.<br>
<br>

- **ApplicationInsights** / **AppInsights**- [Reference to resources Application Insights service]({_APPINSIGHTS_URL})<br>
Application Insights is an extensible Application Performance Management (APM) service for web developers on multiple platforms. Use it to monitor your live web application. It will automatically detect performance anomalies. It includes powerful analytics tools to help you diagnose issues and to understand what users actually do with your app. It's designed to help you continuously improve performance and usability. It works for apps on a wide variety of platforms including .NET, Node.js and J2EE, hosted on-premises or in the cloud. It integrates with your DevOps process, and has connection points to a variety of development tools. It can monitor and analyze telemetry from mobile apps by integrating with Visual Studio App Center.<br>
<br>

- **kqlmagic-readme** - [Reference to Kqlmagic readme]({_KQLMAGIC_README_URL})<br>
<br>

- **kqlmagic-github** - [Reference to Kqlmagic github]({_KQLMAGIC_GITHUB_URL})<br>
<br>

- **kqlmagic-license** - [Reference to Kqlmagic license]({_KQLMAGIC_LICENSE_URL})<br>
<br>

- **kqlmagic-contributors** - [Reference to Kqlmagic contributors]({_KQLMAGIC_CONTRIBUTORS_URL})<br>
<br>

- **kqlmagic-install** - [Reference to Kqlmagic pypi install readme]({_KQLMAGIC_INSTALL_URL})<br>
<br>

- **kqlmagic-downloads** - [Reference to Kqlmagic downloads data]({_KQLMAGIC_DOWNLOADS_URL})<br>
<br>

- **logging** - How to enable logging.<br>
<br>

- **help** - This help.<br>
<br>

{_NEED_SUPPORT_SECTION}"""


_HELP_CONN = f"""## Overview
- To get data from Azure Monitor data resources, the user need to authenticate itself, and if it has the right permission, 
he would be able to query that data resource.
- The current supported data sources are: Azure Data Explorer (kusto) clusters, Application Insights, Log Analytics and Cache.
- Cache data source is not a real data source, it retrieves query results that were cached, but it can only retreive results queries that were executed before, new queries or modified queries won't work.
to get more information on cache data source, execute ```help "cache"```

- The user can connect to multiple data resources.
- Once a connection to a data resource is established, it gets a name of the form <resource>@<data-source>.
- Reference to a data resource can be by connection string, connection name, or current connection (last connection used).
    - If connection is not specified, current connection (last connection used) will be used.
    - To submit queries, at least one connection to a data resource must be established.

- When a connection is specified, and it is a new connection string, the authentication and authorization is validated authomatically, by submiting 
a validation query ```range c from 1 to 10 step 1 | count```, and if the correct result returns, the connection is established.

- An initial connection can be specified as an environment variable.
    - if specified it will be established when Kqlmagic loads.
    - The variable name is ```KQLMAGIC_CONNECTION_STR```

## Authentication methods:

* AAD Username/password - Provide your AAD username and password.
* AAD application - Provide your AAD tenant ID, AAD app ID and app secret.
* AAD code - Provide only your AAD username, and authenticate yourself using a code, generated by ADAL.
* certificate - Provide your AAD tenant ID, AAD app ID, certificate and certificate-thumbprint (supported only with Azure Data Explorer)
* appid/appkey - Provide you application insight appid, and appkey (supported only with Application Insights)
* anonymous - No authentication. For the case that you run your data source locally.

## Connect to Azure Data Explorer (kusto) data resource ```<database or alias>@<cluster>```
Few options to authenticate with Azure Data Explorer (Kusto) data resources:<br>
```%kql azuredataexplorer://code;cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>
```%kql azuredataexplorer://tenant='<tenant-id>';clientid='<aad-appid>';clientsecret='<aad-appkey>';cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>
```%kql azuredataexplorer://tenant='<tenant-id>';certificate='<certificate>';certificate_thumbprint='<thumbprint>';cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>
```%kql azuredataexplorer://tenant='<tenant-id>';certificate_pem_file='<pem_filename>';certificate_thumbprint='<thumbprint>';cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>
```%kql azuredataexplorer://username='<username>';password='<password>';cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>
```%kql azuredataexplorer://anonymous;cluster='<cluster-name>';database='<database-name>';alias='<database-friendly-name>'```<br><br>

Notes:<br>
- username/password works only on corporate network.<br>
- alias is optional.<br>
- if credentials are missing, and a previous connection was established the credentials will be inherited.<br>
- if secret (password / clientsecret / thumbprint) is missing, user will be prompted to provide it.<br>
- if cluster is missing, and a previous connection was established the cluster will be inherited.<br>
- if tenant is missing, and a previous connection was established the tenant will be inherited.<br>
- if only the database change, a new connection can be set as follow: 
```<new-database-name>@<cluster-name>```<br>
- **a not quoted value, is a python expression, that is evaluated and its result is used as the value. This is how you can parametrize the connection string** 

## Connect to Log Analytics data resources ```<workspace or alias>@loganalytics```
Few options to authenticate with Log Analytics:<br>
```%kql loganalytics://code;workspace='<workspace-id>';alias='<workspace-friendly-name>'```<br><br>
```%kql loganalytics://tenant='<tenant-id>';clientid='<aad-appid>';clientsecret='<aad-appkey>';workspace='<workspace-id>';alias='<workspace-friendly-name>'```<br><br>
```%kql loganalytics://username='<username>';password='<password>';workspace='<workspace-id>';alias='<workspace-friendly-name>'```<br><br>
```%kql loganalytics://anonymous;workspace='<workspace-id>';alias='<workspace-friendly-name>'```<br><br>

Notes:<br>
- authentication with appkey works only for the demo.<br>
- username/password works only on corporate network.<br>
- alias is optional.<br>
- if credentials are missing, and a previous connection was established the credentials will be inherited.<br>
- if secret (password / clientsecret) is missing, user will be prompted to provide it.<br>
- if tenant is missing, and a previous connection was established the tenant will be inherited.<br>
- **a not quoted value, is a python expression, that is evaluated and its result is used as the value. This is how you can parametrize the connection string**


## Connect to Application Insights data resources ```<appid or alias>@appinsights```
Few options to authenticate with Apllication Insights:<br><br>
```%kql appinsights://appid='<app-id>';appkey='<app-key>';alias='<appid-friendly-name>'```<br><br>
```%kql appinsights://code;appid='<app-id>';alias='<appid-friendly-name>'```<br><br>
```%kql appinsights://tenant='<tenant-id>';clientid='<aad-appid>';clientsecret='<aad-appkey>';appid='<app-id>';alias='<appid-friendly-name>'```<br><br>
```%kql appinsights://username='<username>';password='<password>';appid='<app-id>';alias='<appid-friendly-name>'```<br><br>
```%kql appinsights://anonymous;appid='<app-id>';alias='<appid-friendly-name>'```<br><br>

Notes:<br>
- username/password works only on corporate network.<br>
- alias is optional.<br>
- if credentials are missing, and a previous connection was established the credentials will be inherited.<br>
- if secret (password / clientsecret / appkey) is missing, user will be prompted to provide it.<br>
- if tenant is missing, and a previous connection was established the tenant will be inherited.<br>
- **a not quoted value, is a python expression, that is evaluated and its result is used as the value. This is how you can parametrize the connection string**

{_NEED_SUPPORT_SECTION}"""


_HELP_CACHE = ""

_HELP_LOGGING = f"""## Overview
- Logging is available, mainly for development debugging. To enable logging one or more of the below environment variable must be set.<br>
The log file is created in the same folder as the current notebook.<br>
<br>

## Logging varaibles
- **{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_LEVEL**<br>
Log level. The following levels are supported: 'FATAL', 'ERROR', 'WARNING', 'INFO', and 'DEBUG'<br>
The default level is: 'DEBUG'<br>
<br>

- **{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE**<br>
Filename for the log messages.<br>
<br>

- **{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_PREFIX**<br>
If filename is not specified, the filename will be build from the prefix as follows: '{{prefix}}-{{ipykernel-unique-key}}.log'<br>
The default preix is: '{Constants.MAGIC_PACKAGE_NAME}'<br>
<br>

- **{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_MODE**<br>
The mode at which the filename is oppened. The following modes are supported: 'append' and 'write'<br>
The default mode is: 'write'<br>
<br>
"""

_HELP_ENV = f"""## Overview
you can configure {Constants.MAGIC_CLASS_NAME} using the following environment variables:

## {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_LEVEL
Log level. The following levels are supported: 'FATAL', 'ERROR', 'WARNING', 'INFO', and 'DEBUG'<br>
Default: 'DEBUG' if a filename is specified, otherwise None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE
Filename for the log messages.<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_PREFIX
If {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE is not specified, the filename will be build from the prefix as follows: '{{prefix}}-{{ipykernel-unique-key}}.log'<br>
Default: '{Constants.MAGIC_PACKAGE_NAME}' if {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_LEVEL is epcified, otherwise None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_MODE
The mode at which the log filename is oppened. The following modes are supported: 'append' and 'write'<br>
Default: 'write' if a filename is specified, otherwise None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_CONNECTION_STR
Set a default connection string. If set, the connection will be set on {Constants.MAGIC_PACKAGE_NAME} load.<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_SSO_ENCRYPTION_KEYS
Specifies details to enable SSO.<br>
Should be a list of key='value' pairs separated by a semicolon ';'<br>
The required keys are: "{SsoEnvVarParam.CACHE_NAME}", "{SsoEnvVarParam.SECRET_KEY}", "{SsoEnvVarParam.SECRET_SALT_UUID}", "{SsoEnvVarParam.CRYPTO}", "{SsoEnvVarParam.STORAGE}"<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_DEVICE_CODE_NOTIFICATION_EMAIL
Specifies details to enable email notification. <br>
Should be a list of key='value' pairs separated by a semicolon ';'<br>
The required keys are: "{Email.SMTP_PORT}", "{Email.SMTP_ENDPOINT}", "{Email.SEND_FROM}", "{Email.SEND_TO}", "{Email.SEND_FROM_PASSWORD}"<br>
The optional keys are: "{Email.CONTEXT}"<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_EXTRAS_REQUIRE
Specifies the same list of extras that were used to install. <br>
if KqlmagicCustom was installed KQLMAGIC_EXTRAS_REQUIRE environment variable should be set to the same list of extras that were used to install (if not set, 'default' will be used)<br>
for example:<br>
KQLMAGIC_EXTRAS_REQUIRE=plotly,pandas,sso<br>
KQLMAGIC_EXTRAS_REQUIRE=default<br>
KQLMAGIC_EXTRAS_REQUIRE=jupyter-basic,widgets<br>
On Kqlmagic load will warn on missing packages that were specified in KQLMAGIC_EXTRAS_REQUIRE (if not set, 'default' will be used)<br>
Features that require a missing dependency will be disabled.<br>
Default: 'default' (all extras)<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_WARN_MISSING_DEPENDENCIES
Specify whether to warn on missing dependencies on {Constants.MAGIC_PACKAGE_NAME} load. Valid values are True or False.<br>
Default: True<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_NOTEBOOK_SERVICE_ADDRESS
Force the specified backend service address. Only for remote kernels.<br>
Default: None (auto set by jupyter implementation)<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_AZUREML_COMPUTE
Specified azureml backend compute host address. If not specified popup windows might not work properly.<br>
Default: value taken from {Constants.MAGIC_CLASS_NAME_UPPER}_NOTEBOOK_SERVICE_ADDRESS<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_NOTEBOOK_APP
Force the specified jupyter implementation. valid values are: "auto", "jupyterlab", "azurenotebook", "azureml", "azuremljupyternotebook", "azuremljupyterlab", "jupyternotebook", "ipython", "visualstudiocode", "azuredatastudio", "nteract", "azuredatastudiosaw"<br>
Default: "auto" (auto detect implementation)<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_LOAD_MODE
If set to 'silent', banner won't be displayed on {Constants.MAGIC_PACKAGE_NAME} load. valid values: 'silent'<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_TEMP_FILES_SERVER_PY_FOLDER 
Force default location of 'my_files_server.py' code. (folder path)<br>
Default: None (auto set by jupyter implementation)<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_TEMP_FILES_SERVER
Specify implementation of file server. valid values are: "auto", "jupyter", "kqlmagic", "disabled"<br>
Default: "auto" (auto set by jupyter implementation) <br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_FILES_SERVER_WINDOW_VISIBILITY
Specify whether to show file server process icon. Valid values are "show" or "hide".<br>
Deafult: "hide"<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_CONFIGURATION
Override default options. Should be a list of key='value' pairs separated by a semicolon ';'<br>
Default: None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_KERNEL
Specify whether {Constants.MAGIC_CLASS_NAME} kernel should be activated. Valid values are True or False.<br>
Default: False<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_CACHE
Specifies a folder name for storing query results that can be used as cache for same queries.<br>
Default:None<br>
<br>

## {Constants.MAGIC_CLASS_NAME_UPPER}_USE_CACHE
Specifies a folder that holds cached query results.<br>
Default: None<br>
<br>
"""


_HELP_SSO = f"""## Overview
- To get data from Azure Monitor data resources, the user need to authenticate itself, and if it has the right permission, 
he would be able to query that data resource.
- It is possible to activate Single Sign On, which will allow a user to go through the authentication process once and remain autheticated for a certain amount of time. 
- To activate Single Sign On:
    1. Set the environmental parameter {Constants.MAGIC_CLASS_NAME_UPPER}_SSO_ENCRYPTION_KEYS with the following parameters:
        - cachename = an identifying name for your SSO cache.
        - secretkey = a password for encryption, for the Single Sign On. Should be at least 8 characters, at least one uppercase Letter, at least 2 digits and at least one non-letter character. Please choose a strong password.
        - secret_salt_uuid = a valid UUID (version 4).
    2. Authenticate normally.
    3. Use this environmental parameter again (with your parameters) to use SSO and authenticate automatically. 
- Additional settings:
    -- Use the option: -enable_sso= False in order to disable SSO- it will not be activated even if correct parameters are given.

{_NEED_SUPPORT_SECTION}"""


_HELP_PROXIES = f"""## Overview
- If you need to use a proxies, configure proxies by setting the environment variables HTTP_PROXY and HTTPS_PROXY.

### http protocol
- Example setting proxies with **http protocol**:<br><br>
```$ export HTTP_PROXY="http://10.10.1.10:3128"```<br><br>
```$ export HTTPS_PROXY="http://10.10.1.10:1080"```<br><br>
<br>
- To use HTTP Basic Auth with your proxy, use the http://user:password@host syntax:<br><br>
```$ export HTTP_PROXY="http://user:pass@10.10.1.10:3128"```<br><br>
```$ export HTTPS_PROXY="http://user:pass@10.10.1.10:1080"```<br><br>
<br>
### socks protocol
- Please make sure to install first requests socks dependency:<br><br>
```$pip install requests[socks]```<br><br>
<br>
- Example setting proxies with **sock protocol**:<br><br>
```$ export HTTP_PROXY="socks5://user:pass@10.10.1.10:3128"```<br><br>
```$ export HTTPS_PROXY="socks5://user:pass@10.10.1.10:1080"```<br><br>
<br>
### Note:
- You can also set the environment variable from within the notebook using %env magic.
"""

_HELP_SAW = f"""## Overview
- {Constants.MAGIC_PACKAGE_NAME} on Secure Access Workstation (SAW) is available, only if it is already preinstalled on the workstation with Azure Data Studio.<br> 
  Azure Data Studio on SAW, expose spython (microsoft secure version of python) and a limited list of preinstalled modules.
  Due to the SAW limitations, {Constants.MAGIC_PACKAGE_NAME} some features are turned off.<br>
<br>
"""

_HELP_SQL = f"""## Overview 
- Azure Data Explorer supports a subset of the SQL language. 
<br>

## Translate SQL to KQL
The primary language to interact with Kusto is KQL (Kusto Query Language). To make the transition and learning experience easier, 
you can use Kusto to translate SQL queries to KQL. Send an SQL query to Kusto, prefixing it with the verb 'EXPLAIN'.
For Example:
```
%%kql
EXPLAIN 
SELECT COUNT_BIG(*) as C FROM StormEvents
```
<br>

## SQL to Kusto cheat sheet
see https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sqlcheatsheet
<br>

## MS-TDS/T-SQL differences between Azure Data Explorer Microsoft SQL Server
see https://docs.microsoft.com/en-us/azure/data-explorer/kusto/api/tds/sqlknownissues
<br>

"""

_HELP_ARIA = f"""## Overview
you can query Aria underlying ADX databases using KQL via Azure Data Explorer
You have to set the connection string with cluster set to: cluster='https://kusto.aria.microsoft.com'; 
and the database to the Aria ADX database name GUID
cluster='https://kusto.aria.microsoft.com';database='<your-GUID';alias='<your-database-friendly-name>'
<br>

## The All-Seeing Eye of Kusto
see https://www.aria.ms/2019/02/01/monthly-release/
<br>

## Aria + Kusto Explorer
see https://www.aria.ms/2018/02/01/introducing-kusto-explorer/
<br>

## ADX Help
see https://www.aria.ms/search/?q=kusto
<br>

"""


_ADX_CLIENT_REQUEST_PROPERTIES = """## Overview

- The client request properties have many uses. 
Some of them are used to make debugging easier (for example, by providing correlation strings that can be used to track client/service interactions), 
others are used to affect what limits and policies get applied to the request.

##List of Client Request Properties

- **deferpartialqueryfailures**: If true, disables reporting partial query failures as part of the result set. [bool]
- **materialized_view_shuffle**: A hint to use shuffle strategy for materialized views that are referenced in the query. The property is an array of materialized views names and the shuffle keys to use. examples: 'dynamic([ { "Name": "V1", "Keys" : [ "K1", "K2" ] } ])' (shuffle view V1 by K1, K2) or 'dynamic([ { "Name": "V1" } ])' (shuffle view V1 by all keys) [dict]
- **max_memory_consumption_per_query_per_node**: Overrides the default maximum amount of memory a whole query may allocate per node. [int]
- **maxmemoryconsumptionperiterator**: Overrides the default maximum amount of memory a query operator may allocate. [int]
- **maxoutputcolumns**: Overrides the default maximum number of columns a query is allowed to produce. [int]
- **norequesttimeout**: Enables setting the request timeout to its maximum value. [bool]
- **notruncation**: Enables suppressing truncation of the query results returned to the caller. [bool]
- **push_selection_through_aggregation**: If true, push simple selection through aggregation [bool]
- **query_bin_auto_at**: When evaluating the bin_auto() function, the start value to use. [LiteralExpression]
- **query_bin_auto_size**: When evaluating the bin_auto() function, the bin size value to use. [LiteralExpression]
- **query_cursor_after_default**: The default parameter value of the cursor_after() function when called without parameters. [str]
- **query_cursor_before_or_at_default**: The default parameter value of the cursor_before_or_at() function when called without parameters. [str]
- **query_cursor_current**: Overrides the cursor value returned by the cursor_current() or current_cursor() functions. [str]
- **query_cursor_disabled**: Disables usage of cursor functions in the context of the query. [bool]
- **query_cursor_scoped_tables**: List of table names that should be scoped to cursor_after_default .. cursor_before_or_at_default (upper bound is optional). [dynamic]
- **query_datascope**: Controls the query's datascope -- whether the query applies to all data or just part of it. ['default', 'all', or 'hotcache']
- **query_datetimescope_column**: Controls the column name for the query's datetime scope (query_datetimescope_to / query_datetimescope_from). [str]
- **query_datetimescope_from**: Controls the query's datetime scope (earliest) -- used as auto-applied filter on query_datetimescope_column only (if defined). [DateTime]
- **query_datetimescope_to**: Controls the query's datetime scope (latest) -- used as auto-applied filter on query_datetimescope_column only (if defined). [DateTime]
- **query_distribution_nodes_span**: If set, controls the way sub-query merge behaves: the executing node will introduce an additional level in the query hierarchy for each sub-group of nodes; the size of the sub-group is set by this option. [Int]
- **query_fanout_nodes_percent**: The percentage of nodes to fanout execution to. [int]
- **query_fanout_threads_percent**: The percentage of threads to fanout execution to. [int]
- **query_force_row_level_security**: If specified, forces Row Level Security rules, even if row_level_security policy is disabled [bool]
- **query_language**: Controls how the query text is to be interpreted. ['csl','kql' or 'sql']
- **query_max_entities_in_union**: Overrides the default maximum number of columns a query is allowed to produce. [int]
- **query_now**: Overrides the datetime value returned by the now(0s) function. [DateTime]
- **query_python_debug**: If set, generate python debug query for the enumerated python node (default first). [bool/int]
- **query_results_apply_getschema** If set, retrieves the schema of each tabular data in the results of the query instead of the data itself. [bool]
- **query_results_cache_max_age**: If positive, controls the maximum age of the cached query results which Kusto is allowed to return [str]
- **query_results_progressive_row_count**: Hint for Kusto as to how many records to send in each update (Takes effect only in progressive mode) [int]
- **query_results_progressive_update_period**: Hint for Kusto as to how often to send progress frames (Takes effect only if in progressive mode) [int]
- **query_take_max_records**: Enables limiting query results to this number of records. [int]
- **queryconsistency**: Controls query consistency. ['strongconsistency' or 'normalconsistency' or 'weakconsistency']
- **request_block_row_level_security**: If specified, blocks access to tables for which row_level_security policy is enabled [bool]
- **request_callout_disabled**: If specified, indicates that the request cannot call-out to a user-provided service. [bool]
- **request_description**: Arbitrary text that the author of the request wants to include as the request description. [str]
- **request_external_table_disabled**: If specified, indicates that the request cannot invoke code in the ExternalTable. [bool]
- **request_impersonation_disabled**: If specified, indicates that the service shouldn't impersonate the caller's identity. [bool]
- **request_readonly**: If specified, indicates that the request must not be able to write anything. [bool]
- **request_remote_entities_disabled**: If specified, indicates that the request cannot access remote databases and clusters. [bool]
- **request_sandboxed_execution_disabled**: If specified, indicates that the request cannot invoke code in the sandbox. [bool]
- **results_progressive_enabled**: If set, enables the progressive query stream
- **servertimeout**: Overrides the default request timeout. [TimeSpan]
- **truncationmaxrecords**: Overrides the default maximum number of records a query is allowed to return to the caller (truncation). [int]
- **truncationmaxsize**: Overrides the dfefault maximum data size a query is allowed to return to the caller (truncation). [int]
- **validate_permissions**: Validates user's permissions to perform the query and doesn't run the query itself. [bool]
"""

_REQUEST_TAGS = f"""## Overview

- Request tags enables to tag **x-ms-app**, **x-ms-user**, **x-ms-client-request-id** and ** ** request headers with a custom string.<br>
The main scenario for tagging request headers is to detect tagged query requests within collected queries telemetry repository.<br>
For example within ADX cluster, executing <br>```.show queries```
<br><br>

##List of Request Tags options

 - **request_id_tag** (idtag) option tags **x-ms-client-request-id** header.<br> 
 The tag will be injected as follows: 
        <br>```x-ms-client-request-id: {Constants.MAGIC_CLASS_NAME}.execute;{{tag}};{{guid}}```<br><br>
    - request_id_tag can be set per query request by setting the the option as follows: <br>```%kql -idtag='{{tag}}' {{query}}```<br>
    - request_id_tag can be set for all requests by setting the default tag: <br>```%config {Constants.MAGIC_CLASS_NAME}.request_id_tag='{{tag}}'```<br>
<br>

 - **request_app_tag** (apptag) option tags **x-ms-app** header.<br>
 The tag will be injected as follows: 
        <br>```x-ms-app: {Constants.MAGIC_CLASS_NAME};{{tag}}```<br><br>
    - request_app_tag can be set per query request by setting the the option as follows: <br>```%kql -apptag='{{tag}}' {{query}}```<br>
    - request_app_tag can be set for all requests by setting the default tag: <br>```%config {Constants.MAGIC_CLASS_NAME}.request_app_tag='{{tag}}'```<br>
 <br>

 - **request_user_tag** (usertag) option tags **x-ms-user** header.<br>
 The tag will be injected as follows: 
        <br>```x-ms-user: {{tag}}```<br><br>
    - request_user_tag can be set per query request by setting the the option as follows: <br>```%kql -usertag='{{tag}}' {{query}}```<br>
    - request_user_tag can be set for all requests by setting the default tag: <br>```%config {Constants.MAGIC_CLASS_NAME}.request_user_tag='{{tag}}'```<br>
<br>
 - **request_user_agent_tag** (uatag) option tags **User-Agent** header.<br>
 The tag will be injected as follows: 
        <br>```User-Agent: {{tag}}```<br><br>
    - request_user_agent_tag can be set per query request by setting the the option as follows: <br>```%kql -uatag='{{tag}}' {{query}}```<br>
    - request_user_agent_tag can be set for all requests by setting the default tag: <br>```%config {Constants.MAGIC_CLASS_NAME}.request_user_agent_tag='{{tag}}'```<br>

"""

_HELP = {
    "query": _KQL_URL,
    "kql": _KQL_URL,
    "appinsights": _APPINSIGHTS_URL, 
    "applicationinsights": _APPINSIGHTS_URL,
    "loganalytics": _LOGANALYTICS_URL,
    "azuremonitor": _AZUREMONITOR_URL,
    "kusto": _KUSTO_URL,
    "adx": _KUSTO_URL,
    "azuredataexplorer": _KUSTO_URL,
    "conn": _HELP_CONN,
    "logging": _HELP_LOGGING,
    "help": _HELP_HELP,
    "usage": _USAGE,
    "commands": _HELP_COMMANDS,
    "cache": _HELP_CACHE,
    "faq": _KQLMAGIC_FAQ_URL,
    "config": "config",
    "options": "options",
    "sso": _HELP_SSO,
    "env": _HELP_ENV, 
    "proxies": _HELP_PROXIES,
    "saw": _HELP_SAW,
    "sql": _HELP_SQL,
    "aria": _HELP_ARIA,
    "clientrequestproperties": _ADX_CLIENT_REQUEST_PROPERTIES,
    "requesttags": _REQUEST_TAGS,
    "kqlmagicdownloads": _KQLMAGIC_DOWNLOADS_URL,
    "kqlmagicinstall": _KQLMAGIC_INSTALL_URL,
    "kqlmagicreadme": _KQLMAGIC_README_URL,
    "kqlmagicgithub": _KQLMAGIC_GITHUB_URL,
    "kqlmagiclicense": _KQLMAGIC_LICENSE_URL,
    "kqlmagiccontributors": _KQLMAGIC_CONTRIBUTORS_URL,
}


class UrlReference(object):
    """ A wrapper class that holds a url reference.
    
    Parameters
    ----------
    name : str
        Name of the url.
    url : str
        Reference url.
    button : str
        A string to be presented on a button, that on click will open the url

     """

    def __init__(self, name:str, url:str, button_text:str, is_raw:bool=None):
        self.name = name
        self.url = url 
        self.button_text = button_text
        self.is_raw = is_raw is True

markdown_module = Dependencies.get_module('markdown', dont_throw=True)
bs4_module = Dependencies.get_module('bs4', dont_throw=True)
lxml_module = Dependencies.get_module('lxml', dont_throw=True)

class MarkdownString(object):
    """ A class that holds a markdown string.
    
    can present the string as markdown, html, and text
     """

    global markdown_module
    global bs4_module
    global lxml_module

    def __init__(self, markdown_string:str, title:str=None):
        self.markdown_string = markdown_string
        self.title = f" - {title}" if title else ""


    # Printable unambiguous presentation of the object
    def __repr__(self)->str:
        return self._markdown_to_text(self.markdown_string)


    def _repr_html_(self)->str:
        if markdown_module:
            return self._force_repr_html_()


    def _force_repr_html_(self)->str:
        return f"""<!DOCTYPE html>
        <html>
        <head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width,initial-scale=1">
        <title>{Constants.MAGIC_PACKAGE_NAME}{self.title}</title>
        </head>
        <body>
        {self._markdown_to_html(self.markdown_string)}
        </body>
        </html>
        """

    def _force_repr_text_(self)->str:
        return self._markdown_to_text(self.markdown_string)


    def _repr_markdown_(self)->str:
        return self.markdown_string
    

    def __str__(self)->str:
        return self.__repr__()


    def _strip_markdown_controls(self, markdown_string:str)->str:
        string = markdown_string.replace("**", "")
        string = string.replace("```", "")
        return string
        

    def _strip_markdown_html_controls(self, markdown_string:str)->str:
        string = markdown_string.replace("<br>", "\n")
        return string


    def _markdown_to_html(self, markdown_string:str)->str:
        if markdown_module:
            html = markdown_module.markdown(markdown_string)
        else:
            string = self._strip_markdown_controls(markdown_string)
            string = string.replace("\n", "<br>")
            html = f"<p>{string}</p>"
        return html

    
    def _markdown_to_text(self, markdown_string:str)->str:
        if markdown_module and bs4_module and lxml_module:
            html = markdown_module.markdown(markdown_string)
            text = ''.join(bs4_module.BeautifulSoup(html, features="lxml").findAll(text=True))
        else:
            text = self._strip_markdown_controls(markdown_string)
            text = self._strip_markdown_html_controls(text)
        return text


def execute_usage_command()->MarkdownString:
    """ execute the usage command.
    command that returns the usage string that will be displayed to the user.

    Returns
    -------
    MarkdownString object
        The usage string wrapped by an object that enable markdown, html or text display.
    """
    return execute_help_command("usage")


def execute_faq_command()->UrlReference:
    """ execute the faq command.
    command that returns the faq string that will be displayed to the user.

    Returns
    -------
    MarkdownString object
        The faq string wrapped by an object that enable markdown, html or text display.
    """
    return execute_help_command("faq")


def execute_help_command(topic:str):
    """ execute the help command.
    command that return the help topic string that will be displayed to the user.

    Returns
    -------
    MarkdownString object
        The help topic string wrapped by an object that enable markdown, html or text display of the topic.
    """
    help_topic_string = _HELP.get(topic.strip().lower().replace("_", "").replace("-", ""))
    if help_topic_string is None:
        raise ValueError(f'"{topic}" unknown help topic')
    elif help_topic_string.startswith("http"):
        button_text = f"popup {topic} reference "
        return UrlReference(topic, help_topic_string, button_text, topic == "faq")
    elif help_topic_string == '':
        help_topic_string = "Sorry, not implemented yet."
    elif help_topic_string in ["config", "options"]:
        return help_topic_string
    return MarkdownString(help_topic_string, title=f"help {topic}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import platform
import uuid



from .parser import Parser
from .display import Display
from .ipython_api import IPythonAPI
from .constants import Constants, CryptoParam, SsoStorageParam, SsoEnvVarParam, SsoStorage, SsoCrypto
from .fernet_crypto import FernetCrypto
from .dpapi_crypto_msal import DpapiCrypto

from .dict_db_storage import DictDbStorage


_SUPPORTED_SSO_STORAGE_TYPES = [
    SsoStorage.IPYTHON_DB
]


_SUPPORTED_SSO_CRYPTO_TYPES = [
    SsoCrypto.DPAPI,
    SsoCrypto.LINUX_LIBSECRET,
    SsoCrypto.OSX_KEYCHAIN,
    SsoCrypto.FERNET
]


class SsoWarning(Exception):
    """Single Sign On Warning class."""

    pass


sso_storage = None


def get_sso_store(cache_selector_key:str=None, **options) -> DictDbStorage:
    global sso_storage
    encryption_keys_string = os.getenv(Constants.SSO_KEYS_ENV_VAR_NAME)
    key_vals = Parser.parse_and_get_kv_string(encryption_keys_string, {}) if encryption_keys_string is not None else {}
    # Display.showWarningMessage(f"Warning: SSO is not activated because environment variable {SSO_KEYS_ENV_VAR_NAME} is not set")

    try:
        crypto_obj = None

        crypto_type = _get_crypto_type(key_vals)

        if crypto_type == SsoCrypto.DPAPI:
            crypto_obj = _create_windows_dpapi_obj()

        elif crypto_type == SsoCrypto.LINUX_LIBSECRET:
            crypto_obj = _create_linux_libsecret_fernet_obj(key_vals)

        elif crypto_type == SsoCrypto.OSX_KEYCHAIN:
            crypto_obj = _create_osx_keychain_fernet_obj(key_vals)

        elif crypto_type == SsoCrypto.FERNET:
            crypto_obj = _create_fernet_obj(key_vals, SsoCrypto.FERNET)

        if crypto_obj is not None:
            cache_name = _get_cache_name(key_vals)
            gc_ttl_in_secs = options.get('sso_db_gc_interval', 0) * Constants.HOUR_SECS  # convert from hours to seconds
            storage_options = {
                SsoStorageParam.CACHE_SELECTOR_KEY: cache_selector_key,
                SsoStorageParam.CRYPTO_OBJ: crypto_obj,
                SsoStorageParam.CACHE_NAME: cache_name,
                SsoStorageParam.GC_TTL_IN_SECS: gc_ttl_in_secs,
            }

            storage_type = _get_storage_type(key_vals, cache_name)
            if storage_type == SsoStorage.IPYTHON_DB:
                db = IPythonAPI._get_ipython_db(**options)
                sso_storage = DictDbStorage(db, storage_options)
                return sso_storage

    except Exception as ex:
        Display.showWarningMessage(f"{ex}")


def clear_sso_store():
    global sso_storage

    storage = sso_storage or get_sso_store()
    if storage is not None:
        storage.clear_db()


def _get_cache_name(key_vals:dict) -> str:
    cache_name = key_vals.get(SsoEnvVarParam.CACHE_NAME, Constants.SSO_DEFAULT_CACHE_NAME)
    return cache_name


def _get_secret_key(key_vals:dict) -> str:
    secret_key = key_vals.get(SsoEnvVarParam.SECRET_KEY) or str(uuid.uuid4())
    return secret_key


def _get_secret_salt_uuid(key_vals:dict) -> str:
    secret_salt_uuid = key_vals.get(SsoEnvVarParam.SECRET_SALT_UUID) or str(uuid.uuid4())
    return secret_salt_uuid


def _get_crypto_type(key_vals:dict) -> str:
    crypto_type = key_vals.get(SsoEnvVarParam.CRYPTO, SsoCrypto.DEFAULT)
    if crypto_type == SsoCrypto.AUTO:
        if platform.system() == 'Linux':
            crypto_type = SsoCrypto.LINUX_LIBSECRET
        elif platform.system() == 'Darwin':
            crypto_type = SsoCrypto.OSX_KEYCHAIN
        elif platform.system() == 'Windows':
            crypto_type = SsoCrypto.DPAPI
        else:
            crypto_type = "unknown"

    _validate_crypto_type(crypto_type)

    return crypto_type


def _validate_crypto_type(crypto_type:str):
    if crypto_type in _SUPPORTED_SSO_CRYPTO_TYPES:
        pass
    elif crypto_type is None:
        raise SsoWarning(f"Warning: SSO is not activated due to environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is missing {SsoEnvVarParam.CRYPTO} key/value")
    else:
        raise SsoWarning(f"Warning: SSO is not activated due to {crypto_type} cryptography is not supported")


def _get_storage_type(key_vals:dict, cache_name:str) -> str:
    storage_type = key_vals.get(SsoEnvVarParam.STORAGE, SsoStorage.DEFAULT)
    _validate_storage_type(storage_type, cache_name)
    return storage_type


def _validate_storage_type(storage_type:str, cache_name:str):
    if storage_type in _SUPPORTED_SSO_STORAGE_TYPES:
        pass
    elif storage_type is None:
        raise SsoWarning(f"Warning: SSO is not activated due to environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is missing {SsoEnvVarParam.STORAGE} key/value")
    else:
        raise SsoWarning(f"Warning: SSO is not activated due to {storage_type} storage is not supported")

    if storage_type == SsoStorage.IPYTHON_DB:
        if not cache_name:
            raise SsoWarning(f"Warning: SSO is not activated due to environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is missing {SsoEnvVarParam.CACHE_NAME} key/value")


def _create_windows_dpapi_obj() -> DpapiCrypto:
    if platform.system() == 'Windows':
        return DpapiCrypto()
    else:
        raise SsoWarning(f"Warning: SSO is not activated due to {SsoCrypto.DPAPI} cryptography is not supported on {platform.system()} platform")

    
def _create_fernet_obj(key_vals:dict, crypto_key:str) -> FernetCrypto:
    if not FernetCrypto.is_installed():
        raise SsoWarning(f"Warning: SSO is not activated due to {SsoCrypto.FERNET} cryptography and/or password-strength modules are not found")

    if crypto_key is None or crypto_key == SsoCrypto.FERNET:
        secret_key = key_vals.get(SsoEnvVarParam.SECRET_KEY) if crypto_key == SsoCrypto.FERNET else _get_secret_key(key_vals)
        if secret_key is None:
            raise SsoWarning(f"Warning: SSO is not activated due to environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is missing {SsoEnvVarParam.SECRET_KEY} key/value")

        secret_salt_uuid = _get_secret_salt_uuid(key_vals)
        if secret_salt_uuid is None:
            raise SsoWarning(f"Warning: SSO is not activated due to environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is missing {SsoEnvVarParam.SECRET_SALT_UUID} key/value")

        hint = FernetCrypto.check_password_strength(secret_key)
        if hint:
            raise SsoWarning(
                f"Warning: SSO could not be activated due to {SsoEnvVarParam.SECRET_KEY} key " 
                f"in environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is too simple. It should contain: \n{hint}")

        try:
            salt = uuid.UUID(secret_salt_uuid, version=4)            
        except:
            raise SsoWarning(
                f"Warning: SSO is not activated due to {SsoEnvVarParam.SECRET_SALT_UUID} key "
                f"in environment variable {Constants.SSO_KEYS_ENV_VAR_NAME} is not set to a valid uuid")

        cache_name = _get_cache_name(key_vals)
        crypto_options = {
            CryptoParam.PASSWORD: f"{cache_name}-{secret_key}",
            CryptoParam.SALT: salt,
            CryptoParam.LENGTH: 32
        }

    else:
        crypto_options = {
            CryptoParam.CRYPTO_KEY: crypto_key,
        }

    return FernetCrypto(crypto_options)


def _create_osx_keychain_fernet_obj(key_vals:dict) -> FernetCrypto:
    if platform.system() == 'Darwin':
        try:
            from msal_extensions.osx import Keychain, KeychainError
        except:
            raise SsoWarning(f"Warning: SSO is not activated due to {SsoCrypto.OSX_KEYCHAIN} cryptography failed to import Keychain, KeychainError from msal_extensions.osx")
        else:
            cache_name = _get_cache_name(key_vals)
            service_name = Constants.MAGIC_CLASS_NAME_LOWER
            account_name = cache_name
            with Keychain() as locker:
                crypto_key = None
                try:
                    crypto_key = locker.get_generic_password(service_name, account_name)
                except KeychainError as ex:
                    if ex.exit_status != KeychainError.ITEM_NOT_FOUND:
                        raise SsoWarning(
                            f"Warning: SSO is not activated due failure to get key from Keychain, {{serviceName: {service_name}, accountName: {account_name}}}. error: {ex}")
                except Exception as ex:
                    raise SsoWarning(
                        f"Warning: SSO is not activated due failure to get key from Keychain, {{serviceName: {service_name}, accountName: {account_name}}}. error: {ex}")
                else:
                    crypto_obj = _create_fernet_obj(key_vals, crypto_key)
                    if crypto_obj is not None and crypto_key is None:
                        try:
                            locker.set_generic_password(service_name, account_name, crypto_obj.crypto_key)
                        except Exception as ex:
                            raise SsoWarning(
                                f"Warning: SSO is not activated due failure to set key to Keychain, {{serviceName: {service_name}, accountName: {account_name}}}. error: {ex}")
                    return crypto_obj

    else:
        raise SsoWarning(f"Warning: SSO is not activated due to {SsoCrypto.OSX_KEYCHAIN} cryptography is not supported on {platform.system()} platform")



def _create_linux_libsecret_fernet_obj(key_vals:dict) -> FernetCrypto:
    if platform.system() == 'Linux':
        try:
            from msal_extensions.libsecret import LibSecretAgent, trial_run
        except:
            raise SsoWarning(
                f"Warning: SSO is not activated due to {SsoCrypto.LINUX_LIBSECRET} cryptography failed to import LibSecretAgent, trial_run from msal_extensions.libsecret")
        else:
            try:
                trial_run()
            except Exception as ex:
                raise SsoWarning(f"""Warning: SSO is not activated due to libsecret did not perform properly.
* If you encountered error "Remote error from secret service:
  org.freedesktop.DBus.Error.ServiceUnknown",
  you may need to install gnome-keyring package.
* Headless mode (such as in an ssh session) is not supported.
error: {ex}
""")
            else:
                cache_name = _get_cache_name(key_vals)
                schema_name = Constants.MAGIC_CLASS_NAME_LOWER
                attributes = {"sso_cache_name": cache_name}
                try:
                    agent = LibSecretAgent(schema_name, attributes)
                    crypto_key = agent.load()
                except Exception as ex:
                    raise SsoWarning(
                        f"Warning: SSO is not activated due failure to load key from Libsecret, {{schemaName: {schema_name}, attributes: {attributes}}}. error: {ex}")
                else:
                    crypto_obj = _create_fernet_obj(key_vals, crypto_key)
                    if crypto_obj is not None and crypto_key is None:
                        try:
                            agent.save(crypto_obj.crypto_key)
                        except Exception as ex:
                            raise SsoWarning(
                                f"Warning: SSO is not activated due failure to save key in Libsecret, {{schemaName: {schema_name}, attributes: {attributes}}}. error: {ex}")
                    return crypto_obj

    else:
        raise SsoWarning(f"Warning: SSO is not activated due to {SsoCrypto.LINUX_LIBSECRET} cryptography is not supported on {platform.system()} platform")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os.path

from wheel.bdist_wheel import bdist_wheel


class azure_bdist_wheel(bdist_wheel):
    """The purpose of this class is to build wheel a little differently than the sdist,
    without requiring to build the wheel from the sdist (i.e. you can build the wheel
    directly from source).
    """

    description = "Create an Azure wheel distribution"

    user_options = bdist_wheel.user_options + [("azure-namespace-package=", None, "Name of the deepest nspkg used")]

    def initialize_options(self):
        bdist_wheel.initialize_options(self)
        self.azure_namespace_package = None

    def finalize_options(self):
        bdist_wheel.finalize_options(self)
        if self.azure_namespace_package and not self.azure_namespace_package.endswith("-nspkg"):
            raise ValueError("azure_namespace_package must finish by -nspkg")

    def run(self):
        if not self.distribution.install_requires:
            self.distribution.install_requires = []
        self.distribution.install_requires.append("{}>=2.0.0".format(self.azure_namespace_package))
        bdist_wheel.run(self)

    def write_record(self, bdist_dir, distinfo_dir):
        if self.azure_namespace_package:
            # Split and remove last part, assuming it's "nspkg"
            subparts = self.azure_namespace_package.split("-")[0:-1]
            folder_with_init = [os.path.join(*subparts[0 : i + 1]) for i in range(len(subparts))]
            for azure_sub_package in folder_with_init:
                init_file = os.path.join(bdist_dir, azure_sub_package, "__init__.py")
                if os.path.isfile(init_file):
                    os.remove(init_file)
                else:
                    raise ValueError("Unable to find {}. Are you sure of your namespace package?".format(init_file))
        bdist_wheel.write_record(self, bdist_dir, distinfo_dir)


cmdclass = {"bdist_wheel": azure_bdist_wheel}


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""A module that allows python comment '#" to prefix %%kql cell
"""

from typing import List, Dict


from traitlets.config.configurable import Configurable


from ._debug_utils import debug_print
from .constants import Constants
from .ipython_api import IPythonAPI


class AllowPyCommentsBeforeCell(object):

    config:Configurable = None
    allow_py_comments_before_cell_default = False

    @staticmethod
    def initialize(config:Configurable):
        AllowPyCommentsBeforeCell.config = config
        AllowPyCommentsBeforeCell.allow_py_comments_before_cell_default = config.allow_py_comments_before_cell
        if config.is_magic is True:
            AllowPyCommentsBeforeCell.allow_py_Comments_before_cell(config.allow_py_comments_before_cell)
        else:
            config.set_trait("allow_py_comments_before_cell", False, force=True)
        
        config.observe(AllowPyCommentsBeforeCell.observe_allow_py_comments_before_cell, names=["allow_py_comments_before_cell"])
        config.observe(AllowPyCommentsBeforeCell.observe_is_magic, names=["is_magic"])


    @staticmethod
    def allow_py_Comments_before_cell(to_activate:bool)->bool:
        to_add_input_transformer = to_activate
        if to_add_input_transformer:
            IPythonAPI.try_add_input_transformers_cleanup(AllowPyCommentsBeforeCell.remove_py_Comments_before_cell)
        else:
            IPythonAPI.try_remove_input_transformers_cleanup(AllowPyCommentsBeforeCell.remove_py_Comments_before_cell)
        return IPythonAPI.is_in_input_transformers_cleanup(AllowPyCommentsBeforeCell.remove_py_Comments_before_cell)


    @staticmethod
    def remove_py_Comments_before_cell(lines:List[str])->List[str]:
        new_lines = lines
        skip_idx = 0
        for line in lines:
            if line.isspace():
                skip_idx += 1
            else:
                first_line = line.lstrip()
                if first_line.startswith(Constants.PYTHON_COMMENT_PREFIX):
                    skip_idx += 1
                else:
                    if first_line.startswith(Constants.CELL_MAGIC_PREFIX):
                        skip_idx += 1
                        new_lines = lines[skip_idx:]
                        new_lines.insert(0, first_line)
                    break
        return new_lines


    @staticmethod
    def observe_allow_py_comments_before_cell(change:Dict[str,str]):
        if change.get('type') == 'change' and change.get('name') == 'allow_py_comments_before_cell':
            if AllowPyCommentsBeforeCell.config.is_magic is False:
                if change.get('new') is True:
                    AllowPyCommentsBeforeCell.config.set_trait('allow_py_comments_before_cell', False, force=True)
            else:
                AllowPyCommentsBeforeCell.allow_py_Comments_before_cell(change.get('new'))


    @staticmethod
    def observe_is_magic(change:Dict[str,str]):
        if change.get('type') == 'change' and change.get('name') == 'is_magic':
            allow_py_comments_before_cell = change.get('new') and AllowPyCommentsBeforeCell.config.allow_py_comments_before_cell
            AllowPyCommentsBeforeCell.config.set_trait('allow_py_comments_before_cell', allow_py_comments_before_cell, force=True)
            AllowPyCommentsBeforeCell.allow_py_Comments_before_cell(allow_py_comments_before_cell)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import json
from datetime import datetime
import collections

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


import dateutil.parser


from .dependencies import Dependencies


class KqlRow(collectionsAbc.Iterator):

    def __init__(self, row, col_num, **options):
        self.options = options
        self.row = row
        self.column_index = 0
        self.columns_count = col_num


    def __iter__(self):
        self.column_index = 0
        return self


    def __next__(self):
        if self.column_index >= self.columns_count:
            raise StopIteration
        val = self.__getitem__(self.column_index)
        self.column_index = self.column_index + 1
        return val


    def __getitem__(self, key):
        if isinstance(key, slice):
            s = self.row[key]
            return KqlRow(s, len(s), **self.options)
        else:
            return self.row[key]


    def __len__(self):
        return self.columns_count


    def __eq__(self, other):
        if len(other) != self.columns_count:
            return False
        for i in range(self.columns_count):
            s = self.__getitem__(i)
            o = other[i]
            if o != s:
                return False
        return True


    def __str__(self):
        return ", ".join(str(self.__getitem__(i)) for i in range(self.columns_count))


    def __repr__(self):
        return self.row.__repr__()


class KqlRowsIter(collectionsAbc.Iterator):
    """ Iterator over returned rows, limited by size """

    def __init__(self, table, row_num, col_num, **options):
        self.options = options
        self.table = table
        self.row_index = 0
        self.rows_count = row_num
        self.col_num = col_num


    def __iter__(self):
        self.row_index = 0
        self.iter_all_iter = self.table.iter_all()
        return self


    def __next__(self):
        if self.row_index >= self.rows_count:
            raise StopIteration
        self.row_index = self.row_index + 1
        return KqlRow(self.iter_all_iter.__next__(), self.col_num, **self.options)


    def __len__(self):
        return self.rows_count


class KqlResponse(object):

    # Object constructor
    def __init__(self, response, **options):
        self.json_response = response.json_response
        self.options = options
        self.completion_query_info = response.completion_query_info_results
        self.completion_query_resource_consumption = response.completion_query_resource_consumption_results
        self.dataSetCompletion = response.dataSetCompletion_results
        self.tables = [
            KqlTableResponse(
                t, 
                response.extended_properties.get(t.id, {}),
                **options) 
            for t in response.primary_results]


class KqlTableResponse(object):

    def __init__(self, data_table, extended_properties:dict, **options):
        self.options = options
        self._extended_properties = extended_properties
        self.data_table = data_table
        self.columns_count = self.data_table.columns_count


    def fetchall(self):
        return KqlRowsIter(self.data_table, self.data_table.rows_count, self.data_table.columns_count, **self.options)


    def fetchmany(self, size):
        return KqlRowsIter(self.data_table, min(size, self.data_table.rows_count), self.data_table.columns_count, **self.options)


    def rowcount(self):
        return self.data_table.rows_count


    def colcount(self):
        return self.data_table.columns_count


    def recordscount(self):
        return self.data_table.rows_count


    def ispartial(self):
        return self.data_table.is_partial


    def keys(self):
        return self.data_table.columns_name


    def types(self):
        return self.data_table.columns_type
    

    @property
    def extended_properties(self):
        " returns properties as specified in ExtendedProperties table"
        return self._extended_properties


    @property
    def datafarme_types(self):
        return [self.KQL_TO_DATAFRAME_DATA_TYPES.get(t) for t in self.data_table.columns_type]


    def _map_columns_to_index(self, columns: list):
        map = {}
        for idx, col in enumerate(columns):
            map[col["ColumnName"]] = idx
        return map


    def returns_rows(self):
        return self.data_table.rows_count > 0


    def to_dataframe(self, raise_errors=True, options=None):
        """Returns Pandas data frame."""
        
        options = options or {}
        pandas = Dependencies.get_module("pandas")

        if self.data_table.columns_count == 0 or self.data_table.rows_count == 0:
            # return pandas.DataFrame()
            pass

        frame = pandas.DataFrame(self.data_table.rows, columns=self.data_table.columns_name)

        for (idx, col_name) in enumerate(self.data_table.columns_name):
            col_type = self.data_table.columns_type[idx].lower()

            if col_type == "timespan":
                frame[col_name] = pandas.to_timedelta(
                    frame[col_name].apply(lambda t: t.replace(".", " days ") if type(t) is str and "." in t.split(":")[0] else t)
                )

            elif col_type == "datetime":
                frame[col_name] = pandas.to_datetime(
                    frame[col_name].apply(lambda d: d if self._is_valid_datetime(d) else None)
                )

            elif col_type == "string":
                # frame[col_name] = frame[col_name].apply(lambda x: json_dumps(x) if type(x) == str else x)
                pass

            elif col_type == "dynamic":
                if options.get("dynamic_to_dataframe") == "str":
                    frame[col_name] = frame[col_name].apply(lambda x: self._dynamic_to_str(x))
                else:
                    frame[col_name] = frame[col_name].apply(lambda x: self._dynamic_to_object(x))

            elif col_type in self.KQL_TO_DATAFRAME_DATA_TYPES:
                pandas_type = self.KQL_TO_DATAFRAME_DATA_TYPES[col_type]
                # NA type promotion
                if pandas_type == "int64" or pandas_type == "int32":
                    for i in range(0, len(frame[col_name])):
                        if frame[col_name][i] is None or str(frame[col_name][i]) == "nan":
                            pandas_type = "float64"
                            break
                elif pandas_type == "bool":
                    for i in range(0, len(frame[col_name])):
                        if frame[col_name][i] is None or str(frame[col_name][i]) == "nan":
                            pandas_type = "object"
                            break
                frame[col_name] = frame[col_name].astype(pandas_type, errors="raise" if raise_errors else "ignore")
        return frame


    def _is_valid_datetime(self, d:str)->bool:
        # max diff in seconds from 9223372036 from 1970-01-01T00:00:00Z
        MAX_DIFF_FROM_EPOCH_IN_SECS = 9223372036  # 2**63/1000000000
        START_EPOCH_DATETIME = dateutil.parser.isoparse('1970-01-01T00:00:00Z')
        try:
            d = dateutil.parser.isoparse(d) if type(d) == str else d
            return isinstance(d, datetime) and abs((d - START_EPOCH_DATETIME).total_seconds()) < MAX_DIFF_FROM_EPOCH_IN_SECS

        except:
            return False


    @staticmethod
    def _dynamic_to_object(value):
        try:
            return json.loads(value) if value and isinstance(value, str) else value
        except Exception:
            return value


    @staticmethod
    def _dynamic_to_str(value):
        try:
            return value if isinstance(value, str) else f"{value}" if value is not None else None
        except Exception:
            return value


    # index MUST be lowercase 
    KQL_TO_DATAFRAME_DATA_TYPES = {
        "bool": "bool",
        "uint8": "int64",
        "int16": "int64",
        "uint16": "int64",
        "int": "int64",
        "uint": "int64",
        "long": "int64",
        "ulong": "int64",
        "float": "float64",
        "real": "float64",
        "decimal": "float64",
        "string": "object",
        "datetime": "datetime64[ns]",
        "guid": "object",
        "timespan": "timedelta64[ns]",
        "dynamic": "object",
        # Support V1
        # "datetime": "datetime64[ns]",
        "int32": "int32",
        "int64": "int64",
        "double": "float64",
        # "string": "object",
        "sbyte": "object",
        # "guid": "object",
        # "timespan": "object",
    }


class FakeResultProxy(object):
    """A fake class that pretends to behave like the ResultProxy.
    """

    # Object constructor
    def __init__(self, cursor, headers):
        self.fetchall = cursor.fetchall
        self.fetchmany = cursor.fetchmany
        self.rowcount = cursor.rowcount
        self.keys = lambda: headers
        self.returns_rows = True



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


import sys
import os
import time
import subprocess as sub
import threading
from typing import Tuple


from .dependencies import Dependencies
from .log import logger
from .constants import Constants
from .my_utils import get_env_var, double_quote
from .ipython_api import IPythonAPI
from .my_utils import quote_spaced_items_in_path


DEFAULT_PROTOCOL = "http"
DEFAULT_HOST = "127.0.0.1"
DEFAULT_PORT = "5000"


class FilesServerManagement(object):

    def __init__(self, server_py_code_path, server_url, base_folder, folders, options=None):
        options = options or {}
        protocol, host, port = self.parse_server_url(server_url)
        self._server_py_code_path = server_py_code_path
        self._protocol = protocol or DEFAULT_PROTOCOL
        self._host = host or DEFAULT_HOST
        self._port = port or self.get_notused_port(host=self._host) or DEFAULT_PORT
        self._base_folder = base_folder
        self._folders = folders  # comma separated folders
        self._server_url = f"{self._protocol}://{self._host}:{self._port}"
        self._is_registered = False
        self._is_started = False
        self._heartbeat_thread = None
        self._kernel_id = options.get("kernel_id")
        self._pid = os.getpid()
        self._liveness_mode = self._find_liveness_mode(sys.platform)

        logger().debug(
            f"FilesServerManagement::startServer init: " 
            f"server_py_code_path: {server_py_code_path}, server_url: {self._server_url}, base_folder: {base_folder}, folders: {folders}, _kernel_id: {self._kernel_id}, liveness_mode: {self._liveness_mode}")

    def get_notused_port(self, host:str="")->str:
        try:
            import socket
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind((host,0))
            s.listen(1)
            port = s.getsockname()[1]
            s.close()
            return str(port)
        except:
            pass


    def parse_server_url(self, url:str)->Tuple[str, str, str]:
        protocol = host = port = None
        if url is not None:
            url = url.lower()
            parts = url.split("://", 1)
            if len(parts) == 2:
                protocol = parts[0]
                url = parts[1]
            parts = url.split(":", 1)
            if len(parts) == 2:
                host = parts[0] if len(parts[0]) > 0 else None
                port = parts[1] if len(parts[1]) > 0 else None
            else:
                try:
                    port = str(int(parts[0]))
                except:
                    host = parts[0]
        return protocol, host, port


    @property
    def files_url(self)->str:
        return f"{self._server_url}/files"


    @property
    def folders_url(self)->str:
        return f"{self._server_url}/folders"


    @property
    def server_url(self)->str:
        return self._server_url




    def startServer(self)->None:
        logger().debug(f"FilesServerManagement::startServer")
        if not self._is_started or not self.pingServer():
            python_exe = sys.executable or 'python'
            os.environ['FLASK_ENV'] = "development"
            window_visibility = get_env_var(f'{Constants.MAGIC_CLASS_NAME_UPPER}_FILES_SERVER_WINDOW_VISIBILITY')
            show_window = window_visibility is not None and window_visibility.lower() == "show"
            show_window = show_window or self._liveness_mode == "show"
            logger().debug(f"FilesServerManagement::startServer with show_window: {show_window}")

            # must use double quotes for base folder and folders, to allow spaces (note single quoted does not work)
            command_params = (
                f'-protocol={self._protocol} -host={self._host} -port={self._port} '
                f'-base_folder={double_quote(self._base_folder)} -folders={double_quote(self._folders)} '
                f'-parent_id={double_quote(self._pid)} -parent_kernel_id={double_quote(self._kernel_id)} '
                f'-liveness_mode={double_quote(self._liveness_mode)} -clean="folders"')
            command_head = "start /min /wait" if show_window else ""
            python_exe = quote_spaced_items_in_path(python_exe)
            command = f'{command_head} {python_exe} {double_quote(self._server_py_code_path)} {command_params}'
            sub.Popen(command, shell=True)
            logger().debug(f"FilesServerManagement::startServer start sub process command: {command}")

            self._is_started = True

            if self._liveness_mode == "heartbeat":
                # alternative mechanism to detect parent is still alive
                # parent is send heartbeat, and if for a period of time heartbeat stops, mean parent exited
                self._heartbeat_thread = HeartbeatThread(self)
                self._heartbeat_thread.start()

        if not self._is_registered:
            IPythonAPI.try_register_to_ipython_atexit(AbortFileServer.abortServer, self._server_url, self._kernel_id, self._heartbeat_thread)
            self._is_registered = True
            logger().debug(f"FilesServerManagement::startServer try_register_to_ipython_atexit: _abortServer({self._server_url}, {self._kernel_id})")


    def _find_liveness_mode(self, platform:str)->str:
        psutil = Dependencies.get_module("psutil")
        requests = Dependencies.get_module("requests")
        liveness_mode = "show"
        if platform == 'win32':
            if psutil:
                liveness_mode = "parent_process_state"
            elif requests:
                liveness_mode = "heartbeat"
        else:
            if self._pid != 1:
                liveness_mode = "parent_process_id_value"
            elif requests:
                liveness_mode = "heartbeat"
            elif psutil:
                liveness_mode = "parent_process_state"

        return liveness_mode


    def pingServer(self)->bool:
        ping_url = None
        result = False
        try:
            ping_url = f'{self._server_url}/ping?kernelid={self._kernel_id}'
            requests = Dependencies.get_module("requests", dont_throw=True)
            if requests:
                requests.get(ping_url)
            result = True
        except:
            pass

        logger().debug(f"FilesServerManagement::pingServer: url: {ping_url}, result: {result}")
        return result


    def heartbeat(self)->bool:
        heartbeat_url = None
        result = False
        try:
            heartbeat_url = f"{self._server_url}/heartbeat?kernelid={self._kernel_id}"
            requests = Dependencies.get_module("requests", dont_throw=True)
            if requests:
                requests.get(heartbeat_url)
                result = True
        except:
            pass

        logger().debug(f"FilesServerManagement::heartbeat: url: {heartbeat_url}, result: {result}")
        return result


    def abortServer(self)->None:
        AbortFileServer.abortServer(self._server_url, self._kernel_id, self._heartbeat_thread)



class HeartbeatThread(threading.Thread):

    def __init__(self, filesServerManagement:FilesServerManagement)->None:
        super(HeartbeatThread, self).__init__()

        self.daemon = True
        self._files_server_management = filesServerManagement
        self._stop_event = threading.Event()


    def run(self)->None:
        while True:
            try:
                if self.stopped():
                    logger().debug(f"HeartbeatThread::HeartbeatThread: exit")
                    break
                self._files_server_management.heartbeat()
            except:
                pass
            time.sleep(1.0)


    def stop(self)->None:
        self._stop_event.set()


    def stopped(self)->bool:
        return self._stop_event.is_set()


class AbortFileServer(object):
    
    @staticmethod
    def abortServer(server_url:str, kernel_id:str, heartbeat_thread:HeartbeatThread)->None:
        abort_url = None
        result = False
        try:
            if heartbeat_thread is not None:
                try:
                    heartbeat_thread.stop()
                except:
                    pass
            abort_url = f"{server_url}/abort?kernelid={kernel_id}"
            requests = Dependencies.get_module("requests", dont_throw=True)
            if requests:
                requests.get(abort_url)
                result = True
        except:
            pass

        logger().debug(f"FilesServerManagement::_abortServer: url: {abort_url}, result: {result}")

        if heartbeat_thread is not None:
            try:
                heartbeat_thread.join()
            except:
                pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""A module to execute that allows %%kql be one line only
"""

from typing import List, Dict


from traitlets.config.configurable import Configurable


from ._debug_utils import debug_print
from .constants import Constants
from .ipython_api import IPythonAPI


class AllowSingleLineCell(object):

    config:Configurable = None
    allow_single_line_cell_default = False

    @staticmethod
    def initialize(config:Configurable):
        AllowSingleLineCell.config = config
        AllowSingleLineCell.allow_single_line_cell_default = config.allow_single_line_cell
        if config.is_magic is True:
            AllowSingleLineCell.allow_single_line_cell(config.allow_single_line_cell)
        else:
            config.set_trait("allow_single_line_cell", False, force=True)
        
        config.observe(AllowSingleLineCell.observe_allow_single_line_cell, names=["allow_single_line_cell"])
        config.observe(AllowSingleLineCell.observe_is_magic, names=["is_magic"])


    @staticmethod
    def allow_single_line_cell(to_activate:bool)->bool:
        to_add_input_transformer = to_activate
        if to_add_input_transformer:
            IPythonAPI.try_add_input_transformers_cleanup(AllowSingleLineCell.add_empty_line_suffix)
        else:
            IPythonAPI.try_remove_input_transformers_cleanup(AllowSingleLineCell.add_empty_line_suffix)
        return IPythonAPI.is_in_input_transformers_cleanup(AllowSingleLineCell.add_empty_line_suffix)


    @staticmethod
    def add_empty_line_suffix(lines:List[str])->List[str]:
        new_lines = lines
        if len(lines) == 1:
            first_line = lines[0]
            first_line_tokens = first_line.split(None, 1)
            if len(first_line_tokens) > 0:
                cell_magic_name = first_line_tokens[0]
                if cell_magic_name in [Constants.CELL_MAGIC_PREFIX]:
                    if not first_line.endswith("\n"):
                        first_line = first_line + "\n"
                    new_lines = [first_line, "\n"]
        return new_lines


    @staticmethod
    def observe_allow_single_line_cell(change:Dict[str,str]):
        if change.get('type') == 'change' and change.get('name') == 'allow_single_line_cell':
            if AllowSingleLineCell.config.is_magic is False:
                if change.get('new') is True:
                    AllowSingleLineCell.config.set_trait('allow_single_line_cell', False, force=True)
            else:
                AllowSingleLineCell.allow_single_line_cell(change.get('new'))


    @staticmethod
    def observe_is_magic(change:Dict[str,str]):
        if change.get('type') == 'change' and change.get('name') == 'is_magic':
            allow_single_line_cell = change.get('new') and AllowSingleLineCell.config.allow_single_line_cell
            AllowSingleLineCell.config.set_trait('allow_single_line_cell', allow_single_line_cell, force=True)
            AllowSingleLineCell.allow_single_line_cell(allow_single_line_cell)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# --------------------------------------------------------------------------
# because it is also executed from setup.py, make sure
# that it imports only modules, that for sure will exist at setup.py execution
# TRY TO KEEP IMPORTS TO MINIMUM
# --------------------------------------------------------------------------

import sys
from setuptools.command.test import test as TestCommand


class TestClass(TestCommand):

    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        try:
            from multiprocessing import cpu_count
            self.pytest_args = ['-n', str(cpu_count()), '--boxed']
        except (ImportError, NotImplementedError):
            self.pytest_args = ['-n', '1', '--boxed']


    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True


    def run_tests(self):
        try:
            import pytest # type: ignore reportMissingImports # pylint: disable=import-error
        except: 
            pass
        else:
            errno = pytest.main(self.pytest_args)
            sys.exit(errno)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, List, Dict, Union, Tuple


from .kql_response import KqlQueryResponse, KqlSchemaResponse
from .kql_proxy import KqlResponse
from .kql_client import KqlClient


class Engine(object):

    # Object constructor
    def __init__(self):
        pass


    @classmethod
    def get_alt_uri_schema_names(cls):
        raise NotImplementedError(cls.__class__.__name__ + ".get_alt_uri_schema_names")


    @classmethod
    def get_reserved_cluster_name(cls):
        raise NotImplementedError(cls.__class__.__name__ + ".get_reserved_cluster_name")


    @classmethod
    def get_mandatory_key(cls)->str:
        raise NotImplementedError(cls.__class__.__name__ + ".get_mandatory_key")


    @classmethod
    def get_valid_keys_combinations(cls)->List[List[str]]:
        raise NotImplementedError(cls.__class__.__name__ + ".get_valid_keys_combinations")


    @classmethod
    def get_uri_schema_name(cls):
        raise NotImplementedError(cls.__class__.__name__ + ".get_uri_schema_name")


    def obfuscate_parsed_conn(self)->Dict[str,str]:
        raise NotImplementedError(self.__class__.__name__ + ".obfuscate_parsed_conn")


    # collect datails, in case bug report will be generated
    def get_details(self)->Dict[str,Any]:
        raise NotImplementedError(self.__class__.__name__ + ".get_details")


    def __eq__(self, other)->bool:
        raise NotImplementedError(self.__class__.__name__ + ".__eq__")


    def get_alias(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_alias")


    def get_database_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_database_name")


    def get_client_database_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_client_database_name")


    def get_cluster_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_cluster_name")


    def get_cluster_friendly_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_cluster_friendly_name")


    def get_database_friendly_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_database_friendly_name")


    def get_id(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_id")


    def get_conn_name(self)->str:
        raise NotImplementedError(self.__class__.__name__ + ".get_conn_name")


    def get_deep_link(self, query:str, options:Dict[str,Any]=None)->str:
        options = options or {}
        raise NotImplementedError(self.__class__.__name__ + ".get_deep_link")


    def createDatabaseFriendlyName(self, dname:str)->str:
        raise NotImplementedError(self.__class__.__name__ + ".createDatabaseFriendlyName")


    def createClusterFriendlyName(self, cname:str)->str:
        raise NotImplementedError(self.__class__.__name__ + ".createClusterFriendlyName")


    def get_client(self)->KqlClient:
        raise NotImplementedError(self.__class__.__name__ + ".get_client")


    def client_execute(self, query:str, user_namespace:Dict[str,Any]=None, database=None, **options)->Union[KqlQueryResponse, KqlSchemaResponse]:
        raise NotImplementedError(self.__class__.__name__ + ".client_execute")


    def execute(self, query:str, user_namespace:Dict[str,Any]=None, database=None, **options)->KqlResponse:
        raise NotImplementedError(self.__class__.__name__ + ".execute")


    def validate(self, **options)->None:
        raise NotImplementedError(self.__class__.__name__ + ".validate")
        

    def validate_database_name(self, **options)->None:
        raise NotImplementedError(self.__class__.__name__ + ".validate_database_name")


    def _get_databases_by_pretty_name(self, **options)->Tuple[List[str],Dict[str,str]]:
        raise NotImplementedError(self.__class__.__name__ + "._get_databases_by_pretty_name")


    def _parse_common_connection_str(self, conn_str:str, current, uri_schema_name:str, mandatory_key:str, keys_combinations:List[str], user_ns:Dict[str,Any])->Dict[str,str]:

        raise NotImplementedError(self.__class__.__name__ + "._parse_common_connection_str")
